from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from PyQt6.QtCore import pyqtSignal
from PyQt6.QtGui import QAction, QCloseEvent
from PyQt6.QtWidgets import QVBoxLayout, QWidget
from PyQt6 import sip
import PyQt6Ads as qtads

from pyrpoc.displays.base_display import BaseDisplay
from pyrpoc.displays.display_registry import display_registry
from pyrpoc.gui.main_widgets.acquisition_mgr import AcquisitionManagerWidget
from pyrpoc.gui.main_widgets.display_mgr import DisplayManagerWidget
from pyrpoc.gui.main_widgets.instrument_mgr import InstrumentManagerWidget
from pyrpoc.gui.main_widgets.menubar import MainMenuBar
from pyrpoc.gui.main_widgets.opto_control_mgr import OptoControlManagerWidget
from pyrpoc.gui.styles.theme_manager import ThemeController
from pyrpoc.services.display_service import DisplayService
from pyrpoc.services.instrument_service import InstrumentService
from pyrpoc.services.modality_service import ModalityService
from pyrpoc.services.opto_control_service import OptoControlService


qtads.CDockManager.setConfigFlag(qtads.CDockManager.eConfigFlag.DisableTabTextEliding, True)
qtads.CDockManager.setConfigFlag(qtads.CDockManager.eConfigFlag.OpaqueSplitterResize, False)


class DockKey(str, Enum):
    ACQUISITION = "acquisition"
    INSTRUMENTS = "instruments"
    DISPLAYS = "displays"
    OPTOCONTROLS = "optocontrols"


@dataclass(frozen=True)
class DockSpec:
    key: DockKey
    title: str
    area: qtads.DockWidgetArea


class MainGUI(QWidget):
    closing = pyqtSignal()

    def __init__(
        self,
        instrument_service: InstrumentService,
        modality_service: ModalityService,
        display_service: DisplayService,
        opto_control_service: OptoControlService,
        theme_controller: ThemeController,
    ):
        super().__init__()
        self.setWindowTitle("pyrpoc")

        self.instrument_service = instrument_service
        self.modality_service = modality_service
        self.display_service = display_service
        self.opto_control_service = opto_control_service
        self.theme_controller = theme_controller

        self.dock_manager = qtads.CDockManager(self)
        self.dock_manager.setStyleSheet("")
        self.docks: list[qtads.CDockWidget] = []
        self.dock_by_key: dict[DockKey, qtads.CDockWidget] = {}
        self.display_docks: dict[BaseDisplay, qtads.CDockWidget] = {}
        self.display_dock_actions: dict[BaseDisplay, QAction] = {}
        self.menubar = MainMenuBar(self)
        self._build_default_docks()

        self.display_service.display_added.connect(self._on_display_added)
        self.display_service.display_removed.connect(self._on_display_removed)
        self.display_service.display_changed.connect(self._on_display_changed)

        layout = QVBoxLayout(self)
        layout.setMenuBar(self.menubar)
        layout.addWidget(self.dock_manager)

        self._refresh_view_menu()
        selected_mode = self.theme_controller.get_saved_mode()
        self.menubar.populate_style_menu(selected_mode)
        self.menubar.style_selected.connect(self.set_style)

    def _build_default_docks(self) -> None:
        dock_specs = [
            DockSpec(DockKey.ACQUISITION, "Acquisition Manager", qtads.DockWidgetArea.LeftDockWidgetArea),
            DockSpec(DockKey.INSTRUMENTS, "Instrument Manager", qtads.DockWidgetArea.LeftDockWidgetArea),
            DockSpec(DockKey.DISPLAYS, "Display Manager", qtads.DockWidgetArea.LeftDockWidgetArea),
            DockSpec(DockKey.OPTOCONTROLS, "Opto-Control Manager", qtads.DockWidgetArea.LeftDockWidgetArea),
        ]

        dock_acq = self.add_dock(
            dock_specs[0].title,
            AcquisitionManagerWidget(self.modality_service),
            dock_specs[0].area,
        )
        dock_acq.setObjectName("dock.acquisition")
        self.dock_by_key[dock_specs[0].key] = dock_acq

        dock_instruments = self.add_dock(
            dock_specs[1].title,
            InstrumentManagerWidget(self.instrument_service),
            dock_specs[1].area,
            tab_with=dock_acq,
        )
        dock_instruments.setObjectName("dock.instruments")
        self.dock_by_key[dock_specs[1].key] = dock_instruments

        dock_displays = self.add_dock(
            dock_specs[2].title,
            DisplayManagerWidget(self.display_service, self.modality_service),
            dock_specs[2].area,
            tab_with=dock_acq,
        )
        dock_displays.setObjectName("dock.displays")
        self.dock_by_key[dock_specs[2].key] = dock_displays

        dock_opto = self.add_dock(
            dock_specs[3].title,
            OptoControlManagerWidget(
                self.opto_control_service,
                self.modality_service,
                self.display_service,
            ),
            dock_specs[3].area,
            tab_with=dock_acq,
        )
        dock_opto.setObjectName("dock.optocontrols")
        self.dock_by_key[dock_specs[3].key] = dock_opto

    def add_dock(
        self,
        title: str,
        widget: QWidget,
        area: qtads.DockWidgetArea,
        tab_with: qtads.CDockWidget | None = None,
    ) -> qtads.CDockWidget:
        dock = qtads.CDockWidget(title)
        dock.setWidget(widget)
        self.docks.append(dock)

        if tab_with is None:
            self.dock_manager.addDockWidget(area, dock)
        else:
            self.dock_manager.addDockWidgetTab(area, dock)

        return dock

    def set_style(self, theme_mode: str) -> None:
        applied = self.theme_controller.apply(theme_mode)
        self.menubar.set_active_style(applied)

    def _refresh_view_menu(self) -> None:
        for display in list(self.display_dock_actions):
            action = self.display_dock_actions.get(display)
            if action is None or sip.isdeleted(action):
                self.display_dock_actions.pop(display, None)
                continue
        self.menubar.populate_view_menu(
            [dock for dock in self.dock_by_key.values()],
            list(self.display_dock_actions.values()),
        )

    def _display_title(self, display: BaseDisplay) -> str:
        name = getattr(display, "user_label", None)
        if name:
            return name
        try:
            cls = display_registry.get_class(display.type_key)
            return getattr(cls, "DISPLAY_NAME", display.type_key)
        except Exception:
            return display.type_key

    def _on_display_added(self, state: object) -> None:
        if not isinstance(state, BaseDisplay):
            return
        if state in self.display_docks:
            self._on_display_changed(state)
            return

        dock = qtads.CDockWidget(self._display_title(state))
        dock.setObjectName(self._display_dock_object_name(state))
        dock.setWidget(state)
        self.display_docks[state] = dock

        try:
            self.dock_manager.addDockWidget(qtads.DockWidgetArea.RightDockWidgetArea, dock)
        except Exception:
            try:
                self.dock_manager.addDockWidget(qtads.DockWidgetArea.LeftDockWidgetArea, dock)
            except Exception:
                dock.deleteLater()
                self.display_docks.pop(state, None)
                return

        action = QAction(self._display_title(state), self)
        action.setCheckable(True)
        visible = bool(getattr(state, "docked_visible", True))
        action.setChecked(visible)
        action.toggled.connect(lambda checked, display=state: self._on_display_dock_toggled(display, checked))
        self.display_dock_actions[state] = action

        if not visible:
            self._set_display_visibility(state, False, update_action=False)
        else:
            self._set_display_visibility(state, True, update_action=False)

        if hasattr(dock, "closed"):
            dock.closed.connect(lambda *_args, display=state: self._on_display_dock_closed(display))

        self._refresh_view_menu()

    def _on_display_removed(self, state: object) -> None:
        if not isinstance(state, BaseDisplay):
            return
        dock = self.display_docks.pop(state, None)
        action = self.display_dock_actions.pop(state, None)

        if action is not None:
            if sip.isdeleted(action):
                action = None
            else:
                try:
                    action.toggled.disconnect()
                except Exception:
                    pass
                try:
                    self.menubar.view_menu.removeAction(action)
                except Exception:
                    pass
                action.setParent(None)
                action.deleteLater()

        if dock is not None:
            if sip.isdeleted(dock):
                dock = None

            if dock is not None:
                detached_widget: QWidget | None = None
                try:
                    if hasattr(self.dock_manager, "removeDockWidget"):
                        self.dock_manager.removeDockWidget(dock)
                except Exception:
                    pass
                try:
                    detached_widget = dock.takeWidget()
                except Exception:
                    detached_widget = None
                if detached_widget is not None:
                    try:
                        detached_widget.setParent(None)
                    except Exception:
                        pass
                dock.deleteLater()

        self._refresh_view_menu()

    def _display_dock_object_name(self, display: BaseDisplay) -> str:
        instance_id = str(getattr(display, "instance_id", "")).strip()
        if not instance_id:
            instance_id = str(id(display))
        safe = "".join(ch if ch.isalnum() or ch in {"_", "-"} else "_" for ch in instance_id)
        return f"dock.display.{safe}"

    def _on_display_changed(self, state: object) -> None:
        if not isinstance(state, BaseDisplay):
            return
        dock = self.display_docks.get(state)
        action = self.display_dock_actions.get(state)
        if dock is not None and sip.isdeleted(dock):
            self.display_docks.pop(state, None)
            dock = None
        if action is not None and sip.isdeleted(action):
            self.display_dock_actions.pop(state, None)
            action = None

        name = self._display_title(state)
        if dock is not None:
            try:
                dock.setWindowTitle(name)
            except Exception:
                pass
        if action is not None:
            action.setText(name)
            if action.isChecked() != bool(getattr(state, "docked_visible", True)):
                action.blockSignals(True)
                action.setChecked(bool(getattr(state, "docked_visible", True)))
                action.blockSignals(False)

        if dock is not None:
            desired_visible = bool(getattr(state, "docked_visible", True))
            current_visible = bool(dock.toggleViewAction().isChecked())
            if current_visible != desired_visible:
                self._set_display_visibility(state, desired_visible, update_action=False)
        self._refresh_view_menu()

    def _display_active(self, display: BaseDisplay) -> bool:
        return display in self.display_service.app_state.displays

    def _on_display_dock_toggled(self, display: BaseDisplay, visible: bool) -> None:
        if not self._display_active(display):
            return
        if bool(getattr(display, "docked_visible", True)) == bool(visible):
            return
        self.display_service.set_dock_visibility(display, visible)

    def _on_display_dock_closed(self, display: BaseDisplay) -> None:
        if not self._display_active(display):
            return
        try:
            self.display_service.detach(display)
        except Exception:
            pass
        try:
            self.display_service.set_dock_visibility(display, False)
        except Exception:
            self._set_display_visibility(display, False)

    def _set_display_visibility(self, display: BaseDisplay, visible: bool, update_action: bool = True) -> None:
        dock = self.display_docks.get(display)
        if dock is None or sip.isdeleted(dock):
            if dock is not None:
                self.display_docks.pop(display, None)
            return

        if dock is not None:
            dock.toggleView(visible)
        action = self.display_dock_actions.get(display)
        if action is not None and sip.isdeleted(action):
            self.display_dock_actions.pop(display, None)
            action = None
        if action is not None and update_action:
            if action.isChecked() != visible:
                action.blockSignals(True)
                action.setChecked(visible)
                action.blockSignals(False)

    def closeEvent(self, event: QCloseEvent) -> None:
        self.closing.emit()
        super().closeEvent(event)



"""In this example, we compare the behavior of histogramming measurement classes (Correlation, Histogram,
HistogramLogBins, and StartStop). We use SynchronizedMeasurements to ensure that all parallel
measurements use the very same subset of time tags.
Besides, the example shows the capabilities of the Time Tagger engine. Some of the histograms have 2 million bins.
The visualization is only getting slow because matplotlib is reaching its limits here."""

from matplotlib import pyplot as plt
import numpy as np

from Swabian import TimeTagger

# Create a TimeTagger instance to control your hardware
tagger = TimeTagger.createTimeTagger()

# Enable the test signal on channel 1 and channel 2 and delay channel 2 by 2 ns
tagger.setTestSignal([1, 2], True)
tagger.setInputDelay(1, 2000)

# Create a SynchronizedMeasurements instance that allows you to control all child
synchronized = TimeTagger.SynchronizedMeasurements(tagger)

# To create synchronized measurement instances, we need a special TimeTagger object
# that we can obtain by the getTagger method. The returned proxy object is not identical
# with the "tagger" object we created above. If we pass the "sync_tagger_proxy" object to a
# measurement, the measurement does NOT start immediately on creation.
# Instead, the "synchronized" object takes control
sync_tagger_proxy = synchronized.getTagger()

# Create a Correlation object. We double the amount of n_bins because Correlation
# provides results for both, positive and negative values
correlation = TimeTagger.Correlation(tagger=sync_tagger_proxy,
                                     channel_1=1,
                                     channel_2=2,
                                     binwidth=1,
                                     n_bins=4E6)
# Histogram is very similar to Correlation, but with fixed roles of start and click channels,
# instead of both channels playing both roles.
histogram = TimeTagger.Histogram(tagger=sync_tagger_proxy,
                                 click_channel=1,
                                 start_channel=2,
                                 binwidth=1,
                                 n_bins=10**6)
# StartStop is a little bit different: firstly, it is single-start/single-stop, so with the
# periodic test signal, we obtain only a single peak. Secondly, it has no predefined
# set of bins and not even a defined number of bins, only the binwidth is fixed. It creates
# bins as they are needed and return pairs of (x, y) values.
start_stop = TimeTagger.StartStop(tagger=sync_tagger_proxy,
                                  click_channel=1,
                                  start_channel=2,
                                  binwidth=1)
# HistogramLogBins is similar to Histogram, but not with a linearly distributed set of equally
# sized bins, but with bins growing as 10^exp_start
histogram_log = TimeTagger.HistogramLogBins(tagger=sync_tagger_proxy,
                                            click_channel=1,
                                            start_channel=2,
                                            exp_start=-12,
                                            exp_stop=-5,
                                            n_bins=400)

print("Acquire the histograms for 5 seconds.")
# We let the synchronized measurements run for 5 s
synchronized.startFor(int(5E12))
synchronized.waitUntilFinished()

# Create the plot
plt.figure(figsize=(9, 6))
histogram_log_edges = histogram_log.getBinEdges()
histogram_log_data = histogram_log.getDataObject().getG2()
np.nan_to_num(histogram_log_data, copy=False)
plt.bar(x=histogram_log_edges[:-1],
        height=histogram_log_data,
        width=np.diff(histogram_log_edges),
        align="edge",
        color=(1, 0, 0),
        edgecolor=(0, 0, 0),
        label="HistogramLogBins",
        zorder=1)
corr_index = correlation.getIndex()
corr_data = correlation.getData()
plt.plot(corr_index,
         corr_data,
         linewidth=3,
         label="Correlation",
         zorder=2)
plt.plot(histogram.getIndex(),
         histogram.getData(),
         linestyle="--",
         label="Histogram",
         zorder=3)
start_stop_data = start_stop.getData()
plt.scatter(start_stop_data[:, 0],
            start_stop_data[:, 1],
            color="green",
            marker=".",
            label="StartStop",
            zorder=4)
plt.xlabel("Time (ps)")
plt.ylabel("Normalized correlation $g^{2}$")
plt.legend(loc='upper right')

# Add some annotations to illustrate the graph
peak = np.argmax(corr_data)
neg_sidepeak = np.argmax(corr_data[:int(.8*2e6)])
pos_sidepeak = -np.argmax(corr_data[:int(-.8*2e6):-1])
plt.annotate("Zoom x axis here:\nAll results are\nvery similar",
             (corr_index[peak], corr_data[peak]/2),
             xytext=(100, 0),
             textcoords="offset pixels",
             arrowprops={"arrowstyle": "->"})
plt.annotate("Zoom here:\nOnly\nCorrelation\nprovides\nneg. data",
             (corr_index[neg_sidepeak], corr_data[neg_sidepeak]),
             xytext=(0, 100),
             textcoords="offset pixels",
             horizontalalignment='center',
             verticalalignment='bottom',
             arrowprops={"arrowstyle": "->"})
plt.annotate("Zoom here:\nStartStop does not\nsee the second peak",
             (corr_index[pos_sidepeak], corr_data[pos_sidepeak]),
             xytext=(0, 100),
             textcoords="offset pixels",
             arrowprops={"arrowstyle": "->"})
last_filled_log_bin = -np.argmax(histogram_log_data[::-1] > 0) - 1
plt.annotate("Zoom here:\nHistogramLogBins\nshows bins for\nlarge delays",
             ((histogram_log_edges[last_filled_log_bin-1] + histogram_log_edges[last_filled_log_bin])/2,
                 histogram_log_data[last_filled_log_bin]),
             xytext=(-100, 100),
             textcoords="offset pixels",
             horizontalalignment='center',
             verticalalignment='bottom',
             arrowprops={"arrowstyle": "->"})
plt.show()

TimeTagger.freeTimeTagger(tagger)



from __future__ import annotations

import numpy as np

import nidaqmx as nx
from nidaqmx.constants import AcquisitionType

from pyrpoc.backend_utils.opto_control_contexts import MaskContext
from pyrpoc.instruments.confocal_daq import ConfocalDAQInstrument
from pyrpoc.optocontrols.base_optocontrol import BaseOptoControl
from pyrpoc.optocontrols.mask import MaskOptoControl


class DaqUnavailableError(RuntimeError):
    """Raised when a DAQ-backed acquisition cannot run on this machine."""


# TODO: make the need for a context object uniform to all optocontrol types
# so that it is easier to know in different modalities what the mask info is shipped as
# this should be enforced by BaseOptoControl
# and we can have a Context object
# this might be a good way to ship DisplayContexts and InstrumentContexts as well

# TODO: make sure that everything needed for acquisition parameter is read into the acquisition object
# and verified at a specific layer enforced by base
# so that we dont need a bunch of validation logic in acquisition code

# TODO: make sure that saving of things is done in acquisition layer
# so that modality service stays agnostic

def extract_mask_contexts(opto_controls: list[BaseOptoControl]) -> list[MaskContext]:
    contexts: list[MaskContext] = []
    for control in opto_controls:
        if not isinstance(control, MaskOptoControl):
            continue

        context = control.context
        if not isinstance(context, MaskContext):
            raise TypeError(f"{type(control).__name__} must prepare a MaskContext before acquisition")

        contexts.append(context)
    return contexts


def generate_raster_waveform(
    total_x: int,
    y_pixels: int,
    pixel_samples: int,
    fast_axis_offset: float,
    fast_axis_amplitude: float,
    slow_axis_offset: float,
    slow_axis_amplitude: float,
) -> np.ndarray:
    fast_amp = max(float(fast_axis_amplitude), 1e-6)
    slow_amp = max(float(slow_axis_amplitude), 1e-6)

    fast_axis = (np.linspace(-1.0, 1.0, total_x, endpoint=False, dtype=np.float32) * fast_amp) + float(fast_axis_offset)
    slow_axis = (np.linspace(-1.0, 1.0, y_pixels, endpoint=False, dtype=np.float32) * slow_amp) + float(slow_axis_offset)

    fast_samples = np.repeat(fast_axis, pixel_samples)
    slow_samples = np.repeat(slow_axis, total_x * pixel_samples)
    fast_raster = np.tile(fast_samples, y_pixels)
    slow_raster = np.tile(slow_samples, 1)

    return np.vstack((fast_raster, slow_raster)).astype(np.float64)


def generate_mask_ttl_signals(
    *,
    total_x: int,
    total_y: int,
    pixel_samples: int,
    extra_left: int,
    extra_right: int,
    device_name: str,
    mask_contexts: list[MaskContext],
    scan_x_pixels: int,
) -> dict[str, np.ndarray]:
    ttl_signals: dict[str, np.ndarray] = {}
    keep_w = max(0, min(scan_x_pixels, max(0, total_x - extra_left - extra_right)))

    for context in mask_contexts:
        if context.mask is None:
            continue

        mask = np.asarray(context.mask, dtype=np.uint8)  # pyright:ignore
        mask_bool = mask > 0
        if mask_bool.size == 0:
            continue

        padded_mask = np.zeros((total_y, total_x), dtype=bool)
        if mask_bool.size:
            source_h = min(mask_bool.shape[0], total_y)
            source_w = min(mask_bool.shape[1], keep_w)
            if source_h > 0 and source_w > 0:
                padded_mask[:source_h, extra_left : extra_left + source_w] = mask_bool[:source_h, :source_w]

        ttl = np.zeros((total_y, total_x, pixel_samples), dtype=bool)
        ttl[padded_mask] = True
        ttl_signals[
            f"{device_name}/port{int(context.daq_port)}/line{int(context.daq_line)}"
        ] = ttl.reshape(-1)
    return ttl_signals


def generate_mask_ttl_signals_split(
    *,
    total_x: int,
    total_y: int,
    pixel_samples: int,
    extra_left: int,
    extra_right: int,
    device_name: str,
    mask_contexts: list[MaskContext],
    scan_x_pixels: int,
    t0_samples: int,
) -> dict[str, np.ndarray]:
    ttl_signals = generate_mask_ttl_signals(
        total_x=total_x,
        total_y=total_y,
        pixel_samples=pixel_samples,
        extra_left=extra_left,
        extra_right=extra_right,
        device_name=device_name,
        mask_contexts=mask_contexts,
        scan_x_pixels=scan_x_pixels,
    )

    if t0_samples >= pixel_samples:
        return ttl_signals

    for channel_name, ttl in ttl_signals.items():
        ttl = ttl.reshape(total_y, total_x, pixel_samples)
        ttl[:, :, t0_samples:] = False
        ttl_signals[channel_name] = ttl.reshape(-1)
    return ttl_signals


def acquire_daq(
    daq_instrument: ConfocalDAQInstrument,
    waveform: np.ndarray,
    ttl_signals: dict[str, np.ndarray],
    x_pixels: int,
    y_pixels: int,
    extra_left: int,
    extra_right: int,
    dwell_time_us: float,
    active_ai_channels: list[int],
) -> tuple[np.ndarray, int, int, int]:
    device_name: str = daq_instrument.device_name
    sample_rate_hz: float = float(daq_instrument.sample_rate_hz)
    fast_axis_channel: int = int(daq_instrument.fast_axis_ao)
    slow_axis_channel: int = int(daq_instrument.slow_axis_ao)

    pixel_samples = max(1, int(dwell_time_us * 1e-6 * sample_rate_hz))
    total_x = x_pixels + extra_left + extra_right
    total_y = y_pixels
    total_samples = total_x * total_y * pixel_samples
    row_samples = total_x * pixel_samples
    keep_start = extra_left * pixel_samples
    keep_end = row_samples - (extra_right * pixel_samples)
    expected_kept = x_pixels * pixel_samples

    waveform_array = np.asarray(waveform, dtype=np.float64)
    ai_channels = [f"{device_name}/ai{idx}" for idx in active_ai_channels]

    do_task: nx.Task | None = None
    static_do_task: nx.Task | None = None
    static_values: list[bool] = []

    try:
        with nx.Task() as ao_task, nx.Task() as ai_task:
            ao_task.ao_channels.add_ao_voltage_chan(f"{device_name}/ao{fast_axis_channel}")
            ao_task.ao_channels.add_ao_voltage_chan(f"{device_name}/ao{slow_axis_channel}")
            for channel in ai_channels:
                ai_task.ai_channels.add_ai_voltage_chan(channel)

            ao_task.timing.cfg_samp_clk_timing(
                rate=sample_rate_hz,
                sample_mode=AcquisitionType.FINITE,
                samps_per_chan=total_samples,
            )
            ai_task.timing.cfg_samp_clk_timing(
                rate=sample_rate_hz,
                source=f"/{device_name}/ao/SampleClock",
                sample_mode=AcquisitionType.FINITE,
                samps_per_chan=total_samples,
            )

            if ttl_signals:
                dynamic_channels: list[str] = []
                dynamic_ttls: list[np.ndarray] = []
                static_channels: list[str] = []
                static_values = []

                for channel_name, ttl in ttl_signals.items():
                    if "/port0/" in channel_name.lower():
                        dynamic_channels.append(channel_name)
                        dynamic_ttls.append(ttl)
                    else:
                        static_channels.append(channel_name)
                        static_values.append(bool(ttl.flat[0]))

                if dynamic_channels:
                    do_task = nx.Task()
                    for channel_name in dynamic_channels:
                        do_task.do_channels.add_do_chan(channel_name)
                    do_task.timing.cfg_samp_clk_timing(
                        rate=sample_rate_hz,
                        source=f"/{device_name}/ao/SampleClock",
                        sample_mode=AcquisitionType.FINITE,
                        samps_per_chan=total_samples,
                    )
                    if len(dynamic_channels) == 1:
                        do_task.write(dynamic_ttls[0].tolist(), auto_start=False)  # pyright:ignore
                    else:
                        do_task.write([ttl.tolist() for ttl in dynamic_ttls], auto_start=False)  # pyright:ignore

                if static_channels:
                    static_do_task = nx.Task()
                    for channel_name in static_channels:
                        static_do_task.do_channels.add_do_chan(channel_name)
                    static_do_task.write(static_values, auto_start=True)  # pyright:ignore

            ao_task.write(waveform_array, auto_start=False)  # pyright:ignore
            ai_task.start()
            if do_task is not None:
                do_task.start()
            ao_task.start()

            timeout = total_samples / sample_rate_hz + 5
            ao_task.wait_until_done(timeout=timeout)
            ai_task.wait_until_done(timeout=timeout)
            if do_task is not None:
                do_task.wait_until_done(timeout=timeout)

            acq_data = ai_task.read(number_of_samples_per_channel=total_samples)  # pyright:ignore
            acq_data = np.asarray(acq_data, dtype=np.float32)
            if acq_data.ndim == 1:
                acq_data = acq_data[np.newaxis, :]
            elif acq_data.ndim != 2:
                raise RuntimeError("Unexpected NI-DAQ data shape")

            frame_channels: list[np.ndarray] = []
            for channel_data in acq_data:
                scan_line_samples = np.asarray(channel_data, dtype=np.float32).reshape(total_y, row_samples)
                if scan_line_samples.shape[1] < keep_end:
                    raise RuntimeError("Acquired sample width is smaller than expected")
                kept_samples = scan_line_samples[:, keep_start:keep_end]
                if kept_samples.shape[1] != expected_kept:
                    raise RuntimeError("Acquired scan payload does not match requested scan dimensions")

                frame_channels.append(kept_samples.astype(np.float32, copy=False))

            return np.stack(frame_channels, axis=0).astype(np.float32, copy=False), total_y, x_pixels, pixel_samples
    except Exception as exc:
        print(f"Error during NI-DAQ acquisition ({exc})")
        raise DaqUnavailableError(f"Error during NI-DAQ acquisition ({exc})") from exc
    finally:
        if do_task is not None:
            do_task.close()
        if static_do_task is not None:
            if static_values:
                try:
                    static_do_task.write([not value for value in static_values], auto_start=True)  # pyright:ignore
                except Exception:
                    pass
            static_do_task.close()


def reshape_data(
    scan_data: np.ndarray,
    total_y: int,
    x_pixels: int,
    pixel_samples: int,
) -> np.ndarray:
    frame_channels: list[np.ndarray] = []
    for channel_data in scan_data:
        channel_pixels = np.asarray(channel_data, dtype=np.float32).reshape(total_y, x_pixels, pixel_samples)
        frame_channels.append(channel_pixels.mean(axis=2))
    return np.stack(frame_channels, axis=0).astype(np.float32, copy=False)


def reshape_data_split(
    scan_data: np.ndarray,
    total_y: int,
    x_pixels: int,
    pixel_samples: int,
    t0_samples: int,
    t1_samples: int,
) -> tuple[np.ndarray, np.ndarray]:
    split_point = int(t0_samples)
    second_start = split_point + int(t1_samples)

    split_channels: list[np.ndarray] = []
    raw_channels: list[np.ndarray] = []

    for channel_data in scan_data:
        pixel_data = np.asarray(channel_data, dtype=np.float32).reshape(total_y, x_pixels, pixel_samples)
        raw_channels.append(pixel_data.astype(np.float32, copy=False))

        first_portion = pixel_data[:, :, :split_point].mean(axis=2)
        if second_start < pixel_samples:
            second_portion = pixel_data[:, :, second_start:].mean(axis=2)
        else:
            second_portion = np.zeros_like(first_portion)

        split_channels.append(first_portion.astype(np.float32, copy=False))
        split_channels.append(second_portion.astype(np.float32, copy=False))

    return (
        np.stack(split_channels, axis=0).astype(np.float32, copy=False),
        np.stack(raw_channels, axis=0).astype(np.float32, copy=False),
    )


def acquire_daq_confocal(
    daq_instrument: ConfocalDAQInstrument,
    x_pixels: int,
    y_pixels: int,
    extra_left: int,
    extra_right: int,
    dwell_time_us: float,
    fast_axis_offset: float,
    fast_axis_amplitude: float,
    slow_axis_offset: float,
    slow_axis_amplitude: float,
    active_ai_channels: list[int],
    mask_contexts: list[MaskContext],
) -> np.ndarray:
    sample_rate_hz: float = float(daq_instrument.sample_rate_hz)
    pixel_samples = max(1, int(dwell_time_us * 1e-6 * sample_rate_hz))
    total_x = x_pixels + extra_left + extra_right

    waveform = generate_raster_waveform(
        total_x=total_x,
        y_pixels=y_pixels,
        pixel_samples=pixel_samples,
        fast_axis_offset=fast_axis_offset,
        fast_axis_amplitude=fast_axis_amplitude,
        slow_axis_offset=slow_axis_offset,
        slow_axis_amplitude=slow_axis_amplitude,
    )
    ttl_signals = generate_mask_ttl_signals(
        total_x=total_x,
        total_y=y_pixels,
        pixel_samples=pixel_samples,
        extra_left=extra_left,
        extra_right=extra_right,
        device_name=daq_instrument.device_name,
        mask_contexts=mask_contexts,
        scan_x_pixels=x_pixels,
    )
    scan_data, total_y, x_pixels_scanned, scan_pixel_samples = acquire_daq(
        daq_instrument=daq_instrument,
        waveform=waveform,
        ttl_signals=ttl_signals,
        x_pixels=x_pixels,
        y_pixels=y_pixels,
        extra_left=extra_left,
        extra_right=extra_right,
        dwell_time_us=dwell_time_us,
        active_ai_channels=active_ai_channels,
    )
    return reshape_data(scan_data, total_y, x_pixels_scanned, scan_pixel_samples)


def acquire_daq_split_confocal(
    daq_instrument: ConfocalDAQInstrument,
    x_pixels: int,
    y_pixels: int,
    extra_left: int,
    extra_right: int,
    dwell_time_us: float,
    fast_axis_offset: float,
    fast_axis_amplitude: float,
    slow_axis_offset: float,
    slow_axis_amplitude: float,
    active_ai_channels: list[int],
    t0_samples: int,
    t1_samples: int,
    mask_contexts: list[MaskContext],
) -> tuple[np.ndarray, np.ndarray]:
    sample_rate_hz: float = float(daq_instrument.sample_rate_hz)
    pixel_samples = max(1, int(dwell_time_us * 1e-6 * sample_rate_hz))
    total_x = x_pixels + extra_left + extra_right

    waveform = generate_raster_waveform(
        total_x=total_x,
        y_pixels=y_pixels,
        pixel_samples=pixel_samples,
        fast_axis_offset=fast_axis_offset,
        fast_axis_amplitude=fast_axis_amplitude,
        slow_axis_offset=slow_axis_offset,
        slow_axis_amplitude=slow_axis_amplitude,
    )
    ttl_signals = generate_mask_ttl_signals_split(
        total_x=total_x,
        total_y=y_pixels,
        pixel_samples=pixel_samples,
        extra_left=extra_left,
        extra_right=extra_right,
        device_name=daq_instrument.device_name,
        mask_contexts=mask_contexts,
        scan_x_pixels=x_pixels,
        t0_samples=t0_samples,
    )
    scan_data, total_y, x_pixels_scanned, scan_pixel_samples = acquire_daq(
        daq_instrument=daq_instrument,
        waveform=waveform,
        ttl_signals=ttl_signals,
        x_pixels=x_pixels,
        y_pixels=y_pixels,
        extra_left=extra_left,
        extra_right=extra_right,
        dwell_time_us=dwell_time_us,
        active_ai_channels=active_ai_channels,
    )
    return reshape_data_split(
        scan_data=scan_data,
        total_y=total_y,
        x_pixels=x_pixels_scanned,
        pixel_samples=scan_pixel_samples,
        t0_samples=t0_samples,
        t1_samples=t1_samples,
    )


"""Welcome to the Time Tagger examples!
In this first example we show how a simple, meaningful measurement can be implemented.
We use the built-in test signal within the Time Tagger.
The applied signal is the very same but has a fixed temporal shift depending on the channel used.
We will compensate for this delay and repeat the measurement to verify the compensation."""

import numpy as np

from Swabian import TimeTagger


def get_delay_and_jitter(x, y):
    # Helper method to calculate the mean time difference of a histogram and the standard deviation.
    mean = np.average(x, weights=y)
    std = np.sqrt(np.average((x-mean)**2, weights=y))
    return mean, std


print("Hello World Example\n")
# Now we connect the Time Tagger.
tagger = TimeTagger.createTimeTagger()

# To generate test data without physical inputs, the Time Tagger has a built-in oscillator providing a
# signal at ~800 kHz. It can be set for each channel individually and will disconnect external signals applied to that channel.
# There is a time offset between the test signals on the range of +/- 2 ns.
tagger.setTestSignal(1, True)
tagger.setTestSignal(2, True)

# We can set the trigger level of the input, although it has no effect if we use the built-in test signal.
# Not taking care of the trigger level is a common source of unexpected behavior!
# Learn more about trigger levels in example 2/B
tagger.setTriggerLevel(channel=1, voltage=0.5)
tagger.setTriggerLevel(channel=2, voltage=0.5)

# Now, we create our first measurement, a 'Correlation' histogram between channels 1 and 2 that will show a maximum
# at the specific delay of between the channels. Learn more about creating measurements in example 1/A.
correlation = TimeTagger.Correlation(tagger=tagger,
                                     channel_1=1,
                                     channel_2=2,
                                     binwidth=10,
                                     n_bins=5000)

print("Acquire a 'Correlation' histogram between channel 1 and 2 for 5 seconds.")
# To run the measurement for exactly one second, we use the startFor method.
# Learn more about starting and stopping measurements in example 1/B.
correlation.startFor(capture_duration=int(5E12))
correlation.waitUntilFinished()

# Analysis of the delay and jitter of the acquired histogram. The data analyzed will reflect the channel-channel jitter
# for your respective Time Tagger.
delay, jitter_rms = get_delay_and_jitter(correlation.getIndex(), correlation.getData())
print("The measured delay is {:.1f} ps with an RMS of {:.1f} ps.".format(delay, jitter_rms))

# Compensate the test signal time offset.
tagger.setInputDelay(channel=2, delay=int(round(delay)))
print("The input delay of channel 2 has been set to {:d} to compensate the time offset of the test signal.".format(int(round(delay))))

# By default, startFor clears the measurement, so we reuse our measurement again without a .clear() command.
# Repeat the measurement to confirm that the delay is compensated.
print("\nAcquire another 'Correlation' histogram between channels 1 and 2 for 5 seconds.")
correlation.startFor(capture_duration=int(5E12))
correlation.waitUntilFinished()

delay, jitter_rms = get_delay_and_jitter(correlation.getIndex(), correlation.getData())
print("The new measured delay is {:.1f} ps with an RMS of {:.1f} ps.".format(delay, jitter_rms))

# Close the connection to the Time Tagger.
TimeTagger.freeTimeTagger(tagger)



import matplotlib.pyplot as plt
import numpy as np

from Swabian.TimeTagger import Flim, createTimeTaggerVirtual


def FastLifetimeComputation(histograms, timesteps):
    """
    Please replace this function with your actual lifetime computation algorithm.
    """
    sum_intensity = histograms.sum(axis=1) + 1e-10  # avoid division by zero
    tau_values = (histograms * timesteps).sum(axis=1) / sum_intensity
    return tau_values


data_file = "Pollen_Galvo_c.1.ttbin"

laser = 1
click1 = -2
frame = 4
pixel = 5
bins = 100
binwidth = 125
xDim = 513
yDim = 513

laser_frequency = 80e6
total_frames = 2

print("This is a example not including the binary dump file required for the analysis.")
print("Please contact the support@swabianinstruments.com for further assistance.\n")

print("Initialize the Time Tagger Virtual")
t = createTimeTaggerVirtual(data_file)

print(
    "Shift the laser signal by one period to compensate for the shift due to the conditional filter"
)
t.setInputDelay(laser, int(-1 / laser_frequency * 1e12))


flim = Flim(
    t,
    start_channel=laser,
    click_channel=click1,
    pixel_begin_channel=pixel,
    n_pixels=xDim * yDim,
    n_bins=bins,
    binwidth=binwidth,
    frame_begin_channel=frame,
    finish_after_outputframe=total_frames,
    n_frame_average=1,
)


# for visualization purpose:
vmax_intensity = 3e7
vmax_flim = 4

# Create a figure with 2 subplots
fig, ax = plt.subplots(2, 2, figsize=(10, 12))  # 1 row, 2 columns
img_display = ax[0, 0].imshow(
    np.zeros((yDim, xDim)), cmap="Blues", vmin=0, vmax=vmax_intensity
)
ax[0, 0].set_title("Intensity (Blues)")
ax[0, 0].set_aspect("equal")
ax[0, 0].set_ylabel('Pixel')

# FLIM image with 'Greens'
flim_display = ax[0, 1].imshow(
    np.zeros((yDim, xDim)), cmap="Greens", vmin=0, vmax=vmax_flim
)
ax[0, 1].set_title("FLIM (Greens)")
ax[0, 1].set_aspect("equal")

# Overlay: RGB image (no colormap)
overlay_display = ax[1, 0].imshow(np.zeros((yDim, xDim, 3)), interpolation="none")
ax[1, 0].set_title("Overlay")
ax[1, 0].set_aspect("equal")
ax[1, 0].set_xlabel('Pixel')
ax[1, 0].set_ylabel('Pixel')

zoomed_overlay = ax[1, 1].imshow(np.zeros((yDim, xDim, 3)), interpolation="none")
ax[1, 1].set_title("Zoomed Overlay")
ax[1, 1].set_aspect("equal")
ax[1, 1].set_xlabel('Pixel')
ax[1, 1].set_xlim(340, xDim)
ax[1, 1].set_ylim(460, 290)

# create a rectangle for zoomed overlay inside the main image
rect = plt.Rectangle(
    (340, 290), 173, 173, linewidth=2, edgecolor="white", facecolor="none"
)
ax[1, 0].add_patch(rect)


plt.ion()
plt.tight_layout()
plt.show()

# for visualization purpose:
# slow down replay speed of the TimeTagger to 0.1x
t.run(0.1)


# Main update loop
while plt.fignum_exists(fig.number):
    base_frame = flim.getCurrentFrameEx()
    sums = base_frame.getIntensities()
    frame = sums.reshape(yDim, xDim)

    img_display.set_data(frame)

    # Get histograms for all pixels: shape (yDim * xDim, bins)
    histograms = base_frame.getHistograms()

    # Create time vector for bins
    bin_width_ns = binwidth * 1e-3
    timesteps = np.arange(bins) * bin_width_ns  # (100,)

    tau_values = FastLifetimeComputation(histograms, timesteps)
    tau_image = tau_values.reshape((yDim, xDim))

    # for visualization purpose
    tau_image[tau_image > vmax_flim] = 0  # Clip values above vmax_flim

    flim_display.set_data(tau_image)

    # Normalize to [0, 1] for display
    intensity_norm = np.clip(frame / vmax_intensity, 0, 1)
    tau_norm = np.clip(tau_image / vmax_flim, 0, 1)

    # Construct RGB image: R=0, G=intensity, B=tau
    overlay_rgb = np.zeros((yDim, xDim, 3))
    overlay_rgb[..., 2] = intensity_norm  # Green channel
    overlay_rgb[..., 1] = tau_norm  # Blue channel

    # Update displays
    img_display.set_data(frame)
    flim_display.set_data(tau_image)
    overlay_display.set_data(overlay_rgb)
    zoomed_overlay.set_data(overlay_rgb)

    # Refresh only image artists (faster)
    for a in [ax[0, 0], ax[0, 1], ax[1, 0], ax[1, 0]]:
        a.draw_artist(a.images[0])

    fig.canvas.flush_events()
    fig.canvas.draw()
    image_from_fig = np.array(fig.canvas.renderer.buffer_rgba())
    plt.pause(0.1)



"""In this example, we measure the jitter/resolution of the Time Tagger channels with the built-in test signals
and compare the results to the specifications of the Time Tagger, which is connected.
"""

from matplotlib import pyplot as plt
import numpy as np
import time
from typing import Dict

from Swabian import TimeTagger


def gaussian(x, mu, sigma, A):
    """Scaled Gaussian function for visual comparison of the measured data"""
    return A/(np.sqrt(2*np.pi)*sigma)*np.exp(-0.5*(x-mu)**2/sigma**2)


def synchronized_correlation_measurement(tagger, channels, binwidth=1, bins=int(5e5), duration=int(15e12)):
    """
    For the jitter measurement, we use correlation measurements of the periodic built-in test signal.
    The function initializes multiple correlation measurements between channels.
    To have a simultaneous start, we make use of the Synchronized Measurement class.
    Returns correlation histogram data for each channel combination.
    """
    correlation_measurements = {}
    # Here we use a single channel, which is correlated to all other channels.
    start_channel = channels[0]
    # Use helper class to synchronize measurements
    with TimeTagger.SynchronizedMeasurements(tagger) as sync_meas:
        for stop_channel in channels:
            # When using Synchronizer, correlate only channels within the same device
            if stop_channel - start_channel >= 100:
                start_channel = stop_channel

            # Do not correlate the channel against itself
            if start_channel == stop_channel:
                continue

            corr_channels = (start_channel, stop_channel)
            # Initiate the measurements and register them for the synchronized measurements
            corr = TimeTagger.Correlation(sync_meas.getTagger(), start_channel, stop_channel, binwidth, bins)
            correlation_measurements[corr_channels] = corr

        # Start all measurements simultaneously
        sync_meas.startFor(duration)
        sync_meas.waitUntilFinished()

        # Collect correlation results
        corr_histograms = {}
        for corr_channels, corr in correlation_measurements.items():
            corr_histograms[corr_channels] = (corr.getIndex(), corr.getData())
    return corr_histograms


def warmup(tagger, duration=int(30e12)):
    """
    Function to warm up the Time Tagger until temperatures are stable.
    """
    all_channels = tagger.getChannelList(TimeTagger.ChannelEdge.Rising)
    tagger.setTestSignal(all_channels, True)
    # Create load on all channels to warm up the Time Tagger
    countrate = TimeTagger.Countrate(tagger, all_channels)
    if tagger.getModel() == "Time Tagger 20" or tagger.getSensorData()[0]['FPGA board']['FPGA Temp'] == 0.0:
        warmup_time = duration/1e12
        print("Warming up ... ", end=' ')
        while warmup_time > 0:
            print(f'{warmup_time:0.0f}', end=' ')
            time.sleep(2)
            warmup_time -= 2
        print("Done.")
    else:
        print("For the Time Tagger Ultra and Time Tagger X, we can check for stable temperatures on the board.")
        warmed_up = False
        cnt = 0
        while not warmed_up:
            pcb_temperatures = []
            fpga_temperatures = []
            # Measure 5 times, then check whether the temperatures on the board have been stable. If not, repeat.
            for i in range(5):
                time.sleep(1)
                cnt = cnt + 1
                sensor_data = tagger.getSensorData()
                pcb = sensor_data[0]['FPGA board']['Board Temp #1']
                fpga = sensor_data[0]['FPGA board']['FPGA Temp']
                pcb_temperatures.append(pcb)
                fpga_temperatures.append(fpga)
                print(f"t: {cnt:3d} s, Board temperature: {pcb:2.1f} °C, FPGA temperature: {fpga:2.1f} °C")
            if (np.ptp(pcb_temperatures) < 1) and (np.ptp(fpga_temperatures) < 1.5):
                warmed_up = True
                print("The Time Tagger is warmed up.")
                print("")
            else:
                print("The Time Tagger is not yet warmed up.")
    del countrate


def get_jitter_and_delay(hist_t, hist_c):
    """
    Calculate RMS jitter and mean (giving the internal delay of the test signal to later shift the plots)
    """
    mean_t = np.average(hist_t, weights=hist_c)
    jitter_rms = np.sqrt(np.average((hist_t-mean_t)**2, weights=hist_c))
    return jitter_rms, mean_t


# Searching for available Time Taggers
serials = TimeTagger.scanTimeTagger()
print(f'Found {len(serials)} connected Time Tagger(s)')

# Connect to the Time Tagger and collect device information.
# Please specify the serial number explicitly in case more than one device is connected
# and not already taken care of by a synchronizer.
tagger = TimeTagger.createTimeTagger(serial='')
serial = tagger.getSerial()
model = tagger.getModel()
lic_info = tagger.getDeviceLicense()
edition = lic_info[0]['edition']

resolution_modes = ['Standard']
if edition == 'HighRes':
    if model == 'Time Tagger Ultra':
        resolution_modes.extend(['HighResA', 'HighResB', 'HighResC'])
    elif model == 'Time Tagger X':
        resolution_modes.append('HighResB')
    else:
        raise ValueError('Unknown combination of model and edition!')

print('Time Tagger info: ')
print(' Serial:          ', serial)
print(' Model:           ', model)
print(' Edition:         ', edition)
print(' Resolution modes:', ', '.join(resolution_modes))

if model not in ['Time Tagger 20', 'Time Tagger Ultra', 'Time Tagger X']:
    raise ValueError(f'Currently {model} is not supported by this example script')

time.sleep(0.5)
print('Warming up the Time Tagger to get an accurate jitter measurement.')
warmup(tagger)

# Storing our data in dictionaries to account for the different modes of the HighRes option.
measured_jitters: Dict[str, list] = {}
within_specs: Dict[str, list] = {}
measured_channels: Dict[str, list] = {}
spec_jitter_label = 'Spec. typical jitter'
specified_jitter_rms = {}

# We loop over the different HighRes modes, if available. Or only use the single available mode.
for i, mode in enumerate(resolution_modes):
    measured_jitters[mode] = []
    within_specs[mode] = []
    measured_channels[mode] = []

    print(f'Operating the Time Tagger in {mode} resolution mode')
    # To change the resolution mode, first, we need to disconnect the Time Tagger
    TimeTagger.freeTimeTagger(tagger)
    resolution = getattr(TimeTagger.Resolution, mode)  # get enum value
    tagger = TimeTagger.createTimeTagger(serial=serial, resolution=resolution)

    assert isinstance(tagger, TimeTagger.TimeTagger)

    # Retrieve rms resolution value from the first channel (a constant value from device specifications)
    # Determine channel resolutions specification from the device
    tagger_configuration = tagger.getConfiguration()
    input_cfg = tagger_configuration['inputs'][0]
    specified_jitter_rms[mode] = input_cfg['resolution rms']

    if mode == 'Standard':
        available_inputs = tagger.getChannelList(TimeTagger.ChannelEdge.Rising)
    else:
        # Measure only on HighRes channels
        available_inputs = tagger.getChannelList(TimeTagger.ChannelEdge.HighResRising)

    print(' Available inputs:', available_inputs)

    # Enable test signals on the available inputs
    tagger.setTestSignal(available_inputs, True)

    print('Measuring for 30 seconds...')
    # Retrieving the measured data
    correlation_data = synchronized_correlation_measurement(tagger, available_inputs, duration=int(30e12))

    print('Measurement complete.\nNow evaluating the data')
    time.sleep(1)

    fig, ax = plt.subplots()  # Create a plot to visualize the TWO channel jitter
    ax.set_title('Visual comparison of the measured two-channel jitters to specifications')
    ax.set_xlabel('Time (ps)')
    ax.set_ylabel('kCounts')
    # Looping over the measurement data, evaluating the single channel RMS jitter from it and displaying the results
    # For the visual comparison a Gaussian with standard deviation = sqrt(2)*specified_RMS_jitter is used
    # since we look at the two-channel jitter
    for corr_channels, (hist_t, hist_c) in correlation_data.items():
        start_channel, stop_channel = corr_channels
        rms_t, mean_t = get_jitter_and_delay(hist_t, hist_c)
        jitter_measured = rms_t / np.sqrt(2)  # per channel jitter

        # Plot measure correlation curve
        label = f'{start_channel:>2d} vs {stop_channel:>2d}'
        line, = ax.plot(hist_t-mean_t, hist_c/1e3)

        # Check if measured jitter is within specifications
        jitter_spec = specified_jitter_rms[mode]
        print((f'Correlation {label} Measured channel jitter: {jitter_measured:0.1f} ps ',
               spec_jitter_label,
               f'{jitter_spec: 0.1f} ps'))
        measured_channels[mode].append(label)
        measured_jitters[mode].append(jitter_measured)

    mean_jitter = np.mean(measured_jitters[mode])
    print('\n')
    print(f'mean jitter: {mean_jitter:0.1f} ps, within specs: {mean_jitter < jitter_spec}')
    print('\n')

    line.set_label(label)  # Add only one label to prevent a crowded plot

    # Plot Gaussian function corresponding to specified jitter
    hist_c_model = gaussian(x=hist_t, mu=0, sigma=jitter_spec*np.sqrt(2), A=np.sum(hist_c))
    ax.plot(hist_t, hist_c_model/1e3, color='k', ls='--', label=spec_jitter_label)

    ax.set_xlim(-5*jitter_spec, 5*jitter_spec)
    ax.legend(loc=1)
    plt.show()

print('Plotting a summary of the results')
# Summary of the measured RMS jitters for a single channel in a table

color_map = {True: 'green', False: 'red'}  # Cell colors
colLabels = ['Channels', 'single channel\nRMS jitter (ps)', spec_jitter_label+' (ps)']

for i, mode in enumerate(resolution_modes):
    fig, ax = plt.subplots()
    rowLabels = measured_channels[mode]
    jitters = measured_jitters[mode]
    jitter_spec = specified_jitter_rms[mode]

    cellContent = [(ch, f'{jitt:0.1f}', jitter_spec) for ch, jitt in zip(rowLabels, jitters)]
    the_table = ax.table(cellText=cellContent, colLabels=colLabels, loc='center', cellLoc='center')
    the_table.scale(1, 1.5)
    ax.axis('tight')
    ax.axis('off')
    fig.suptitle(f'{model} {edition} {mode}')
    fig.tight_layout()

# Freeing the Time Tagger again so that it can be used in another application
TimeTagger.freeTimeTagger(tagger)
plt.show()



from __future__ import annotations

from PyQt6.QtWidgets import QWidget

from . import handlers
from pyrpoc.gui.main_widgets.opto_control_mgr.state import OptoControlManagerState
from pyrpoc.gui.main_widgets.opto_control_mgr.ui import build_opto_control_manager_ui
from pyrpoc.services.display_service import DisplayService
from pyrpoc.services.modality_service import ModalityService
from pyrpoc.services.opto_control_service import OptoControlService


class OptoControlManagerWidget(QWidget):
    def __init__(
        self,
        opto_control_service: OptoControlService,
        modality_service: ModalityService,
        display_service: DisplayService,
        parent: QWidget | None = None,
    ):
        super().__init__(parent)
        self.opto_control_service = opto_control_service
        self.modality_service = modality_service
        self.display_service = display_service

        self.state = OptoControlManagerState()
        self.ui = build_opto_control_manager_ui(self)

        self.type_combo = self.ui.type_combo
        self.add_btn = self.ui.add_btn
        self.instances_layout = self.ui.instances_layout

        self._wire_signals()
        self._refresh_available()
        self._refresh_instances()

    def _wire_signals(self) -> None:
        self.add_btn.clicked.connect(self._on_add_clicked)
        self.opto_control_service.inventory_changed.connect(self._refresh_instances)
        self.modality_service.modality_selected.connect(self._on_modality_selected)

    def _selected_type_key(self) -> str | None:
        value = self.type_combo.currentData()
        return str(value) if value else None

    def _refresh_available(self) -> None:
        handlers.refresh_available(self)

    def _refresh_instances(self) -> None:
        '''
        Service inventory changes should only diff cards.
        This preserves expanded card bodies and avoids stale QObject deref on add/remove.
        '''
        handlers.refresh_instances(self)

    def _on_add_clicked(self) -> None:
        handlers.on_add_clicked(self)

    def _on_modality_selected(self, key: str) -> None:
        handlers.on_modality_selected(self, key)



from __future__ import annotations

from typing import Any

import cv2
import numpy as np

from pyrpoc.backend_utils.opto_control_contexts import MaskContext


def generate_toy_confocal_frame(
    *,
    x_pixels: int,
    y_pixels: int,
    active_channels: list[int],
    frame_index: int,
    mask_contexts: list[MaskContext],
    fast_axis_offset: float,
    fast_axis_amplitude: float,
    slow_axis_offset: float,
    slow_axis_amplitude: float,
) -> np.ndarray:
    if x_pixels <= 0 or y_pixels <= 0:
        raise ValueError("Scan dimensions must be positive")

    if not active_channels:
        raise ValueError("No active AI channels configured")

    frame = np.zeros((len(active_channels), y_pixels, x_pixels), dtype=np.float32)
    fast_amp = max(float(fast_axis_amplitude), 1e-6)
    slow_amp = max(float(slow_axis_amplitude), 1e-6)

    for channel_index, ai_channel in enumerate(active_channels):
        channel = _build_toy_channel(
            x_pixels=x_pixels,
            y_pixels=y_pixels,
            frame_index=frame_index,
            channel_index=channel_index,
            ai_channel=ai_channel,
            fast_axis_offset=float(fast_axis_offset),
            fast_axis_amplitude=fast_amp,
            slow_axis_offset=float(slow_axis_offset),
            slow_axis_amplitude=slow_amp,
        )
        frame[channel_index] = channel

    _apply_masks(frame, mask_contexts)
    return frame


def generate_toy_split_confocal_frame(
    *,
    x_pixels: int,
    y_pixels: int,
    active_channels: list[int],
    frame_index: int,
    mask_contexts: list[MaskContext],
    fast_axis_offset: float,
    fast_axis_amplitude: float,
    slow_axis_offset: float,
    slow_axis_amplitude: float,
    t0_samples: int,
    t1_samples: int,
    pixel_samples: int,
) -> tuple[np.ndarray, np.ndarray]:
    if pixel_samples <= 0:
        raise ValueError("pixel_samples must be positive")
    if t0_samples < 1:
        raise ValueError("t0 samples must be >= 1")
    if t1_samples < 0:
        raise ValueError("t1 samples must be >= 0")
    if t0_samples + t1_samples >= pixel_samples:
        raise ValueError(
            f"split timing uses all samples: t0={t0_samples}, t1={t1_samples}, pixel_samples={pixel_samples}"
        )

    frame = generate_toy_confocal_frame(
        x_pixels=x_pixels,
        y_pixels=y_pixels,
        active_channels=active_channels,
        frame_index=frame_index,
        mask_contexts=mask_contexts,
        fast_axis_offset=fast_axis_offset,
        fast_axis_amplitude=fast_axis_amplitude,
        slow_axis_offset=slow_axis_offset,
        slow_axis_amplitude=slow_axis_amplitude,
    )

    rng = np.random.default_rng(seed=(frame_index + 1) * 17)
    raw = np.repeat(frame[:, :, :, None], pixel_samples, axis=3)
    raw = raw + rng.normal(0.0, 0.02, size=raw.shape).astype(np.float32)

    first_start = int(t0_samples)
    second_start = first_start + int(t1_samples)

    first_portion = raw[:, :, :, :first_start].mean(axis=3)
    if second_start < pixel_samples:
        second_portion = raw[:, :, :, second_start:].mean(axis=3)
    else:
        second_portion = np.zeros_like(first_portion)

    first_portion = _overlay_split_label(first_portion, "1")
    second_portion = _overlay_split_label(second_portion, "2")

    split_frame = np.stack((first_portion, second_portion), axis=1).reshape(frame.shape[0] * 2, y_pixels, x_pixels)
    return split_frame.astype(np.float32, copy=False), raw.astype(np.float32, copy=False)


def _build_toy_channel(
    *,
    x_pixels: int,
    y_pixels: int,
    frame_index: int,
    channel_index: int,
    ai_channel: int,
    fast_axis_offset: float,
    fast_axis_amplitude: float,
    slow_axis_offset: float,
    slow_axis_amplitude: float,
) -> np.ndarray:
    seed = (frame_index + 1) * 1009 + (channel_index + 1) * 101 + ai_channel * 17
    rng = np.random.default_rng(seed)
    channel = np.zeros((y_pixels, x_pixels), dtype=np.float32)

    x = np.linspace(-1.0, 1.0, x_pixels, dtype=np.float32)
    y = np.linspace(-1.0, 1.0, y_pixels, dtype=np.float32)
    xx, yy = np.meshgrid(x, y)

    x_term = (xx - fast_axis_offset) / fast_axis_amplitude
    y_term = (yy - slow_axis_offset) / slow_axis_amplitude
    channel += 0.15 * np.sin((x_term + 0.07 * frame_index) * (5.0 + channel_index))
    channel += 0.12 * np.cos((y_term - 0.05 * frame_index) * (4.0 + 0.5 * channel_index))

    n_shapes = int(rng.integers(10, 18))
    for _ in range(n_shapes):
        intensity = float(rng.uniform(0.2, 1.0))
        if rng.random() < 0.5:
            cx = int(rng.integers(0, x_pixels))
            cy = int(rng.integers(0, y_pixels))
            radius = int(rng.integers(max(3, min(x_pixels, y_pixels) // 30), max(6, min(x_pixels, y_pixels) // 8)))
            cv2.circle(
                channel,
                (cx, cy),
                radius,
                intensity,
                thickness=-1,
                lineType=cv2.LINE_AA,
            )
        else:
            cx = int(rng.integers(0, x_pixels))
            cy = int(rng.integers(0, y_pixels))
            a = int(rng.integers(max(4, x_pixels // 40), max(8, x_pixels // 10)))
            b = int(rng.integers(max(4, y_pixels // 40), max(8, y_pixels // 10)))
            angle = float(rng.uniform(0.0, 180.0))
            cv2.ellipse(
                channel,
                (cx, cy),
                (a, b),
                angle,
                0.0,
                360.0,
                intensity,
                thickness=-1,
                lineType=cv2.LINE_AA,
            )

    channel += rng.normal(0.0, 0.03, size=(y_pixels, x_pixels)).astype(np.float32)
    channel -= float(np.min(channel))
    peak = float(np.max(channel))
    if peak > 0:
        channel /= peak

    return channel.astype(np.float32, copy=False)


def _apply_masks(
    frame: np.ndarray,
    mask_contexts: list[MaskContext],
) -> None:
    if frame.ndim != 3:
        return
    _, h, w = frame.shape

    for context in mask_contexts:
        mask = np.asarray(context.mask, dtype=np.uint8) #pyright:ignore
        if mask.ndim != 2:
            continue

        if mask.shape != (h, w):
            mask = cv2.resize(mask, (w, h), interpolation=cv2.INTER_NEAREST)

        active = mask > 0
        if not np.any(active):
            continue

        for idx in range(frame.shape[0]):
            boost = float(np.max(frame[idx]))
            frame[idx, active] += boost


def _overlay_split_label(image: np.ndarray, label: str) -> np.ndarray:
    if image.ndim != 2:
        return image

    if image.shape[0] < 16 or image.shape[1] < 24:
        return image

    marker = np.zeros(image.shape[:2], dtype=np.uint8)
    baseline = 0.8 * image.shape[0] if image.shape[0] > 0 else 0
    x = max(2, int(image.shape[1] * 0.05))
    y = int(max(14, baseline))
    cv2.putText(
        marker,
        label,
        (x, y),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.6,
        255,
        2,
        cv2.LINE_AA,
    )

    if not np.any(marker):
        return image

    weight = 0.2
    return image + weight * (marker.astype(np.float32) / 255.0) * (float(np.max(np.abs(image))) + 1.0)



from __future__ import annotations

from typing import TYPE_CHECKING

from PyQt6.QtWidgets import QMessageBox

from pyrpoc.instruments.base_instrument import BaseInstrument
from pyrpoc.gui.main_widgets.opto_control_mgr.instance_card import InstanceCardWidget

if TYPE_CHECKING:
    from pyrpoc.gui.main_widgets.instrument_mgr.widget import InstrumentManagerWidget


def refresh_available(widget: InstrumentManagerWidget) -> None:
    """Populate dropdown from registered instrument classes.

    Route:
    - widget startup / state refresh
    - -> `InstrumentService.list_available`
    - -> registry descriptors rendered in combo.
    """
    current_key = widget._selected_type_key()
    widget.type_combo.blockSignals(True)
    widget.type_combo.clear()
    for row in widget.instrument_service.list_available():
        key = row["key"]
        name = row.get("display_name", key)
        widget.type_combo.addItem(name, key)
    widget.type_combo.blockSignals(False)

    if current_key:
        idx = widget.type_combo.findData(current_key)
        if idx >= 0:
            widget.type_combo.setCurrentIndex(idx)
    elif widget.type_combo.count() > 0:
        widget.type_combo.setCurrentIndex(0)


def refresh_instances(widget: InstrumentManagerWidget) -> None:
    """Synchronize card widgets with current service inventory.

    Signal refresh flow:
    - service emits `inventory_changed` after add/remove
    - -> this function diffs desired rows against current card map
    - -> only changed cards are created/removed.
    """
    desired_cards: dict[BaseInstrument, InstanceCardWidget] = {}
    rows = widget.instrument_service.list_instances()
    wanted_states = [row["state"] for row in rows]

    _remove_missing_cards(widget, set(wanted_states))

    for row in rows:
        state: BaseInstrument = row["state"]
        name = row["name"]

        card = widget.state.card_widgets.get(state)
        if card is None:
            card = _create_card(widget, state, name, row["key"])
            widget.state.card_widgets[state] = card

        _refresh_card_text(card, state, name, row["key"])
        desired_cards[state] = card

    _reorder_cards(widget, desired_cards, rows)


def on_expand_requested(widget: InstrumentManagerWidget, state_obj: object) -> None:
    """Handle lazy widget creation for a card body.

    Route:
    - card Expand click
    - -> this handler
    - -> `InstrumentService.get_widget`
    - -> card body widget attach.
    """
    if not isinstance(state_obj, BaseInstrument):
        return
    card: InstanceCardWidget = widget.state.card_widgets.get(state_obj)  # pyright: ignore
    if card is None:
        return

    card.set_expanded(not card.is_expanded())
    if not card.is_expanded():
        return

    if card.body_layout.count() != 0:
        if _is_stale_widget(card):
            card.set_body_widget(None)
            _reset_instrument_widget_cache(state_obj)
            _attach_widget_to_card(widget, state_obj, card)
            return
        card.set_local_status("Status: ready")
        return

    if _attach_widget_to_card(widget, state_obj, card):
        return

    # Retry path for stale PyQt objects: clear any stale body child, force widget
    # refresh on the underlying instance, and build again once.
    _reset_instrument_widget_cache(state_obj)
    _attach_widget_to_card(widget, state_obj, card)


def on_add_clicked(widget: InstrumentManagerWidget) -> None:
    """Add selected instrument class into app state.

    Route:
    - Add button click
    - -> this handler
    - -> `InstrumentService.create_instrument`
    - -> inventory signal for card refresh.
    """
    key = widget._selected_type_key()
    if not key:
        return
    try:
        widget.instrument_service.create_instrument(key)
        widget.status_label.setText("Status: added instrument")
    except Exception as exc:
        show_error(widget, str(exc))


def on_remove_requested(widget: InstrumentManagerWidget, state_obj: object) -> None:
    """Remove an instrument instance and persist via service.

    Route:
    - card Remove click
    - -> this handler
    - -> `InstrumentService.remove_instrument`.
    """
    if isinstance(state_obj, BaseInstrument):
        widget.instrument_service.remove_instrument(state_obj)
        widget.status_label.setText("Status: removed instrument")


def show_error(widget: InstrumentManagerWidget, message: str) -> None:
    """Render a modal error and reflect it in the manager status label."""
    widget.status_label.setText(f"Status: error - {message}")
    QMessageBox.critical(widget, "Instrument Error", message)


def _attach_widget_to_card(widget: InstrumentManagerWidget, state_obj: BaseInstrument, card: InstanceCardWidget) -> bool:
    """Build one concrete widget body for an expanded card.

    Returns True if widget could be instantiated and attached.
    """
    try:
        child = widget.instrument_service.get_widget(
            state_obj,
            parent=card.body_container,
            on_change=lambda s=state_obj, c=card, w=widget: _on_widget_changed(w, s, c),
        )
        card.set_body_widget(child)
        card.set_local_status("Status: ready")
        return True
    except Exception as exc:
        # If this fails with a deleted-object style exception, caller may clear
        # cached widget state on the instrument and retry once.
        card.set_local_status(f"Status: error - {exc}")
        return False


def _is_stale_widget(card: InstanceCardWidget) -> bool:
    """Probe body widget liveness before trusting cached wrapper references."""
    if card.body_layout.count() == 0:
        return False
    widget_item = card.body_layout.itemAt(0)
    if widget_item is None:
        return False
    body_widget = widget_item.widget()
    if body_widget is None:
        return False
    try:
        body_widget.isVisible()
    except Exception as exc:  # typically PyQt wrapper-after-delete errors
        _mark_stale_status(card, f"Status: warning - stale body widget detected ({exc})")
        return True
    return False


def _mark_stale_status(card: InstanceCardWidget, status: str) -> None:
    """Attach a diagnostic status to the card while keeping manager flow alive."""
    card.set_local_status(status)


def _reset_instrument_widget_cache(state_obj: BaseInstrument) -> None:
    """Reset cached widget state on the concrete instrument object if present."""
    if hasattr(state_obj, "widget"):
        try:
            setattr(state_obj, "widget", None)
        except Exception:
            pass


def _remove_missing_cards(widget: InstrumentManagerWidget, wanted_states: set[BaseInstrument]) -> None:
    """Delete only cards for states that no longer exist in service inventory."""
    for state, card in list(widget.state.card_widgets.items()):
        if state not in wanted_states:
            widget.state.card_widgets.pop(state)
            card.setParent(None)
            card.deleteLater()


def _create_card(
    widget: InstrumentManagerWidget,
    state: BaseInstrument,
    name: str,
    key: str,
) -> InstanceCardWidget:
    """Create a new card for a new instrument state."""
    card = InstanceCardWidget(state, name, widget)
    card.set_enable_visible(False)
    card.set_marker_text(f"[{key}]")
    card.remove_requested.connect(lambda state_obj, w=widget: on_remove_requested(w, state_obj))
    card.expand_requested.connect(lambda state_obj, w=widget: on_expand_requested(w, state_obj))
    _refresh_card_text(card, state, name, key)
    return card


def _refresh_card_text(card: InstanceCardWidget, state: BaseInstrument, name: str, key: str) -> None:
    """Keep visible card metadata aligned with registry metadata."""
    summary = state.get_collapsed_summary().strip()
    title = f"{name} ({summary})" if summary else name
    card.title_label.setText(title)
    card.set_marker_text(f"[{key}]")


def _on_widget_changed(widget: InstrumentManagerWidget, state_obj: BaseInstrument, card: InstanceCardWidget) -> None:
    widget.status_label.setText("Status: widget changed")
    widget.instrument_service.mark_instance_changed(state_obj)
    key = state_obj.type_key
    row_name = key
    for row in widget.instrument_service.list_available():
        if row["key"] == key:
            row_name = row.get("display_name", key)
            break
    _refresh_card_text(card, state_obj, str(row_name), key)


def _reorder_cards(
    widget: InstrumentManagerWidget,
    desired_cards: dict[BaseInstrument, InstanceCardWidget],
    rows: list[dict[str, object]],
) -> None:
    """Rebuild layout order to match service rows without recreating card instances."""
    desired_order: list[BaseInstrument] = [row["state"] for row in rows]

    # Lift widgets out of layout while preserving instances.
    current_cards: list[InstanceCardWidget] = []
    while widget.instances_layout.count() > 0:
        item = widget.instances_layout.takeAt(0)
        card = item.widget()  # pyright: ignore
        if card is not None:
            current_cards.append(card)

    for state in desired_order:
        card = desired_cards.get(state)
        if card is None:
            continue
        widget.instances_layout.insertWidget(widget.instances_layout.count(), card)



from __future__ import annotations

from typing import Literal

from PyQt6.QtCore import QFile, QSettings, QTextStream
from PyQt6.QtWidgets import QApplication

from . import breeze_all

ThemeMode = Literal['dark', 'light']

_SETTINGS_ORG = 'pyrpoc'
_SETTINGS_APP = 'pyrpoc'
_SETTINGS_KEY_THEME_MODE = 'ui/theme_mode'
AVAILABLE_BREEZE_THEMES = [
    'dark-blue', 
    'dark-blue-alt', 
    'dark-cyan', 
    'dark-cyan-alt', 
    'dark-green', 
    'dark-green-alt', 
    'dark-pink', 
    'dark-pink-alt', 
    'dark-purple', 
    'dark-purple-alt', 
    'dark-red', 
    'dark-red-alt', 
    'light-blue', 
    'light-blue-alt', 
    'light-cyan', 
    'light-cyan-alt', 
    'light-green', 
    'light-green-alt', 
    'light-pink', 
    'light-pink-alt', 
    'light-purple', 
    'light-purple-alt', 
    'light-red', 
    'light-red-alt'
]



class ThemeController:
    AVAILABLE_MODES: tuple[ThemeMode, ...] = ('dark', 'light')

    def __init__(self, app: QApplication):
        self.app = app
        self.settings = QSettings(_SETTINGS_ORG, _SETTINGS_APP)

    def get_available_modes(self) -> list[ThemeMode]:
        return list(self.AVAILABLE_MODES)

    def get_saved_mode(self) -> ThemeMode:
        raw = self.settings.value(_SETTINGS_KEY_THEME_MODE, 'dark')
        mode = str(raw).strip().lower()
        if mode in self.AVAILABLE_MODES:
            return mode
        return 'dark'

    def apply_saved_or_default(self) -> ThemeMode:
        return self.apply(self.get_saved_mode(), persist=False)

    def _load_breeze_stylesheet(self, mode: ThemeMode) -> str:
        qss_path = ':/dark-pink/stylesheet.qss'

        file = QFile(qss_path)
        ok = file.open(QFile.OpenModeFlag.ReadOnly | QFile.OpenModeFlag.Text)
        if not ok:
            raise RuntimeError(f'failed to open breeze stylesheet: {qss_path}')

        stream = QTextStream(file)
        return stream.readAll()

    def apply(self, mode: str, persist: bool = True) -> ThemeMode:
        normalized = mode.strip().lower()
        if normalized not in self.AVAILABLE_MODES:
            normalized = 'dark'
        selected_mode: ThemeMode = normalized

        if persist:
            self.settings.setValue(_SETTINGS_KEY_THEME_MODE, selected_mode)

        base_qss = self._load_breeze_stylesheet(selected_mode)
        self.app.setStyleSheet(base_qss)
        return selected_mode


"""The FileReader allows you to read data from files that have been stored by the FileWriter.
With the FileReader, you get access to the raw time tags which you can use for whatever you
want. If you want to use the highly efficient measurement classes provided by the API, however,
the Virtual Time Tagger (example 3-B) is the best choice.
In this example, we first use the FileWriter to store test signal data for one second. In the
second part, we use the FileReader to load the time tags from the ttbin file."""

import os
import tempfile
from time import sleep

from Swabian import TimeTagger

# Create a TimeTagger instance to control your hardware
tagger = TimeTagger.createTimeTagger()

# Enable the test signal on channels 1 and 2
tagger.setTestSignal([1, 2], True)

# The FileWriter provides a temporary ttbin file for the example
tempdir = tempfile.TemporaryDirectory()
filename = tempdir.name + os.sep + "filewriter.ttbin"
filewriter = TimeTagger.FileWriter(tagger=tagger,
                                   filename=filename,
                                   channels=[1, 2])
sleep(1)
filewriter.stop()
print("\nFileWriter has prepared a ttbin file with {} events.".format(filewriter.getTotalEvents()))
del filewriter

# To read the data form the temporary file, we create a FileReader object
filereader = TimeTagger.FileReader(filename)

print('\nRead back the dumped stream from the filesystem using FileReader.\n')
input("-> Press Enter to show some selected time tags")

# The format for the table and the head of the table
format_string = '{:>8} | {:>17} | {:>7} | {:>14} | {:>13}'
print(format_string.format('TAG #', 'EVENT TYPE', 'CHANNEL', 'TIMESTAMP (ps)', 'MISSED EVENTS'))
print('---------+-------------------+---------+----------------+--------------')

n_events = 100000  # Number of events to read at once
event_name = ['0 (TimeTag)', '1 (Error)', '2 (OverflowBegin)', '3 (OverflowEnd)', '4 (MissedEvents)']
i = 0
while filereader.hasData():
    # getData() does not return timestamps, but an instance of TimeTagStreamBuffer
    # that contains more information than just the timestamp
    data = filereader.getData(n_events=n_events)

    # With the following methods, we can retrieve a numpy array for the particular information:
    channel = data.getChannels()            # The channel numbers
    timestamps = data.getTimestamps()       # The timestamps in ps
    overflow_types = data.getEventTypes()   # TimeTag = 0, Error = 1, OverflowBegin = 2, OverflowEnd = 3, MissedEvents = 4
    missed_events = data.getMissedEvents()  # The numbers of missed events in case of overflow

    # Output to table
    if i < 2 or not filereader.hasData():
        print(format_string.format(*" "*5))
        heading = ' Start of data chunk {} with {} events '.format(i+1, data.size)
        extra_width = 69 - len(heading)
        print('{} {} {}'.format("="*(extra_width//2), heading, "="*(extra_width - extra_width//2)))
        print(format_string.format(*" "*5))
        print(format_string.format(i*n_events + 1, event_name[overflow_types[0]], channel[0], timestamps[0], missed_events[0]))
        if data.size > 1:
            print(format_string.format(i*n_events + 2, event_name[overflow_types[1]], channel[1], timestamps[1], missed_events[1]))
        if data.size > 3:
            print(format_string.format(*["..."]*5))
        if data.size > 2:
            print(format_string.format(i*n_events + data.size, event_name[overflow_types[-1]], channel[-1], timestamps[-1], missed_events[-1]))
    if i == 1:
        print(format_string.format(*" "*5))
        for j in range(3):
            print(format_string.format(*"."*5))
    i += 1

# Close the FileReader and remove the temporary files
del filereader
tempdir.cleanup()

TimeTagger.freeTimeTagger(tagger)



# matplotlib for the plots, including its Qt backend
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT
from matplotlib.figure import Figure
# numpy for statistical analysis
import numpy as np
import os
from PyQt5.QtWidgets import QMainWindow, QApplication, QFileDialog
from PyQt5.QtCore import QTimer
from PyQt5 import uic
import sys

# all required TimeTagger dependencies
from Swabian.TimeTagger import Coincidences, Counter, Correlation, createTimeTagger, freeTimeTagger


class CoincidenceExample(QMainWindow):
    ''' Small example of how to create a UI for the TimeTagger with the PySide2/PyQt5 framework'''

    def __init__(self, tagger):
        '''Constructor of the coincidence example window
        The TimeTagger object must be given as arguments to support running many windows at once.'''

        # Create the UI from the designer file and connect its action buttons
        super(CoincidenceExample, self).__init__()

        # Please use the QtDesigner to edit the ui interface file
        my_dir = os.path.dirname(__file__)
        self.ui = uic.loadUi(os.path.join(my_dir, 'CoincidenceExampleWindow.ui'), self)
        self.ui.startButton.clicked.connect(self.startClicked)
        self.ui.stopButton.clicked.connect(self.stopClicked)
        self.ui.clearButton.clicked.connect(self.clearClicked)
        self.ui.saveButton.clicked.connect(self.saveClicked)

        # Update the measurements whenever any input configuration changes
        self.ui.channelA.valueChanged.connect(self.updateMeasurements)
        self.ui.channelB.valueChanged.connect(self.updateMeasurements)
        self.ui.delayA.valueChanged.connect(
            lambda value: self.setInputDelay('A', value)
        )
        self.ui.delayB.valueChanged.connect(
            lambda value: self.setInputDelay('B', value)
        )
        self.ui.triggerA.valueChanged.connect(
            lambda value: self.setTriggerLevel('A', value)
        )
        self.ui.triggerB.valueChanged.connect(
            lambda value: self.setTriggerLevel('B', value)
        )
        self.ui.testsignalA.stateChanged.connect(
            lambda state: self.setTestSignal('A', state != 0)
        )
        self.ui.testsignalB.stateChanged.connect(
            lambda state: self.setTestSignal('B', state != 0)
        )
        self.ui.coincidenceWindow.valueChanged.connect(self.updateMeasurements)
        self.ui.correlationBinwidth.valueChanged.connect(
            self.updateMeasurements)
        self.ui.correlationBins.valueChanged.connect(self.updateMeasurements)

        # Create the matplotlib figure with its subplots for the counter and correlation
        self.fig = Figure()
        self.counterAxis = self.fig.add_subplot(211)
        self.correlationAxis = self.fig.add_subplot(212)
        self.canvas = FigureCanvasQTAgg(self.fig)
        self.toolbar = NavigationToolbar2QT(self.canvas, self)
        self.ui.plotWidget.layout().addWidget(self.toolbar)
        self.ui.plotWidget.layout().addWidget(self.canvas)

        # Create the TimeTagger measurements
        self.running = True
        self.measurements_dirty = False
        self.tagger = tagger
        self.last_channels = [0, 0]
        self.last_coincidenceWindow = 0
        self.updateMeasurements()

        # Use a timer to redraw the plots every 100ms
        self.timer = QTimer(interval=100, timeout=self.updateCounterPlot)
        self.timer.start()

    def getCouterNormalizationFactor(self):
        bin_index = self.counter.getIndex()
        # normalize 'clicks / bin' to 'kclicks / second'
        return 1e12 / bin_index[1] / 1e3

    def updateMeasurements(self):
        '''Create/Update all TimeTagger measurement objects'''

        # If any configuration is changed while the measurements are stopped, recreate them on the start button
        if not self.running:
            self.measurements_dirty = True
            return

        # Set the input delay, trigger level, and test signal of both channels
        channels = [self.ui.channelA.value(), self.ui.channelB.value()]
        self.tagger.setInputDelay(channels[0], self.ui.delayA.value())
        self.tagger.setInputDelay(channels[1], self.ui.delayB.value())
        self.tagger.setTriggerLevel(channels[0], self.ui.triggerA.value())
        self.tagger.setTriggerLevel(channels[1], self.ui.triggerB.value())
        self.tagger.setTestSignal(channels[0], self.ui.testsignalA.isChecked())
        self.tagger.setTestSignal(channels[1], self.ui.testsignalB.isChecked())

        # Only recreate the counter if its parameter has changed,
        # else we'll clear the count trace too often
        coincidenceWindow = self.ui.coincidenceWindow.value()
        if self.last_channels != channels or self.last_coincidenceWindow != coincidenceWindow:
            self.last_channels = channels
            self.last_coincidenceWindow = coincidenceWindow

            # Create the virtual coincidence channel
            self.coincidences = Coincidences(
                self.tagger,
                [channels],
                coincidenceWindow
            )

            # Measure the count rate of both input channels and the coincidence channel
            # Use 200 * 50ms binning
            self.counter = Counter(
                self.tagger,
                channels + list(self.coincidences.getChannels()),
                binwidth=int(50e9),
                n_values=200
            )

        # Measure the correlation between A and B
        self.correlation = Correlation(
            self.tagger,
            channels[1],
            channels[0],
            self.ui.correlationBinwidth.value(),
            self.ui.correlationBins.value()
        )

        # Create the measurement plots
        self.counterAxis.clear()
        self.plt_counter = self.counterAxis.plot(
            self.counter.getIndex() * 1e-12,
            self.counter.getData().T * self.getCouterNormalizationFactor()
        )
        self.counterAxis.set_xlabel('time (s)')
        self.counterAxis.set_ylabel('count rate (kEvents/s)')
        self.counterAxis.set_title('Count rate')
        self.counterAxis.legend(['A', 'B', 'coincidences'])
        self.counterAxis.grid(True)

        self.correlationAxis.clear()
        index = self.correlation.getIndex()
        data = self.correlation.getDataNormalized()
        self.plt_correlation = self.correlationAxis.plot(
            index * 1e-3,
            data
        )
        self.plt_gauss = self.correlationAxis.plot(
            index * 1e-3,
            data,
            linestyle='--'
        )
        self.correlationAxis.axvspan(
            -coincidenceWindow/1000.,
            coincidenceWindow/1000.,
            color='green',
            alpha=0.3
        )
        self.correlationAxis.set_xlabel('time (ns)')
        self.correlationAxis.set_ylabel('normalized correlation')
        self.correlationAxis.set_title('Correlation between A and B')
        self.correlationAxis.grid(True)

        # Generate nicer plots
        self.fig.tight_layout()

        self.measurements_dirty = False

        # Update the plot with real numbers
        self.updateCounterPlot()

    def getTaggerChannel(self, label):
        """Resolve channel label into the Time Tagger channel number"""
        assert label in 'AB', 'Unknown channel label "{}"'.format(label)
        return int(getattr(self.ui, 'channel{}'.format(label)).value())

    def setInputDelay(self, channel, value):
        """Set input delay on channel A or B"""
        tt_channel = self.getTaggerChannel(channel)
        self.tagger.setInputDelay(tt_channel, value)

    def setTestSignal(self, channel, enable):
        """Enable/Disable test signal on the channel A or B"""
        tt_channel = self.getTaggerChannel(channel)
        self.tagger.setTestSignal(tt_channel, enable)

    def setTriggerLevel(self, channel, value):
        """Set trigger level on channel A or B"""
        tt_channel = self.getTaggerChannel(channel)
        self.tagger.setTriggerLevel(tt_channel, value)

    def startClicked(self):
        '''Handler for the start action button'''
        self.running = True

        if self.measurements_dirty:
            # If any configuration is changed while the measurements are stopped,
            # recreate them on the start button
            self.updateMeasurements()
        else:
            # else manually start them
            self.counter.start()
            self.correlation.start()

    def stopClicked(self):
        '''Handler for the stop action button'''
        self.running = False
        self.counter.stop()
        self.correlation.stop()

    def clearClicked(self):
        '''Handler for the clear action button'''
        self.correlation.clear()

    def saveClicked(self):
        '''Handler for the save action button'''

        # Ask for a filename
        filename, _ = QFileDialog().getSaveFileName(
            parent=self,
            caption='Save to File',
            directory='CoincidenceExampleData.txt',  # default name
            filter='All Files (*);;Text Files (*.txt)',
            options=QFileDialog.DontUseNativeDialog
        )

        # And write all results to disk
        if filename:
            with open(filename, 'w') as f:
                f.write('Input channel A: %d\n' % self.ui.channelA.value())
                f.write('Input channel B: %d\n' % self.ui.channelB.value())
                f.write('Input delay A: %d ps\n' % self.ui.delayA.value())
                f.write('Input delay B: %d ps\n' % self.ui.delayB.value())
                f.write('Trigger level A: %.3f V\n' % self.ui.triggerA.value())
                f.write('Trigger level B: %.3f V\n' % self.ui.triggerB.value())
                f.write('Test signal A: %d\n' %
                        self.ui.testsignalA.isChecked())
                f.write('Test signal B: %d\n' %
                        self.ui.testsignalB.isChecked())

                f.write('Coincidence window: %d ps\n' %
                        self.ui.coincidenceWindow.value())
                f.write('Correlation bin width: %d ps\n' %
                        self.ui.correlationBinwidth.value())
                f.write('Correlation bins: %d\n\n' %
                        self.ui.correlationBins.value())

                f.write('Counter data:\n%s\n\n' %
                        self.counter.getData().__repr__())
                f.write('Correlation data:\n%s\n\n' %
                        self.correlation.getData().__repr__())

    def resizeEvent(self, event):
        '''Handler for the resize events to update the plots'''
        self.fig.tight_layout()
        self.canvas.draw()

    def updateCounterPlot(self):
        '''Handler for the timer event to update the plots'''
        if self.running:
            # Counter
            data = self.counter.getData() * self.getCouterNormalizationFactor()
            for data_line, plt_counter in zip(data, self.plt_counter):
                plt_counter.set_ydata(data_line)
            self.counterAxis.relim()
            self.counterAxis.autoscale_view(True, True, True)

            # Calculate the expectation and the standard deviation of the correlation
            # With this two values, we can display a Gaussian fit
            index = self.correlation.getIndex()
            data = self.correlation.getDataNormalized()
            total = np.sum(data)
            if total > 0:
                offset = np.sum(data * index) / total
                stddev = np.sqrt(np.sum(data * (index - offset)**2) / total)
            else:
                offset = 0
                stddev = 0
            if stddev > 0:
                corr_binwidth = self.ui.correlationBinwidth.value()
                A = corr_binwidth * total / np.sqrt(2*np.pi*stddev**2)
                gauss = A * np.exp(- 0.5 * (index - offset)**2 / stddev**2)
            else:
                gauss = index * 0

            # Correlation
            self.plt_correlation[0].set_ydata(
                self.correlation.getDataNormalized())
            self.plt_gauss[0].set_ydata(gauss)
            self.correlationAxis.relim()
            self.correlationAxis.autoscale_view(True, True, True)
            self.correlationAxis.legend(['measured correlation', '$\\mu$=%.1fps, $\\sigma$=%.1fps' % (
                offset, stddev), 'coincidence window'])
            self.canvas.draw()


# If this file is executed, initialize QApplication, create a TimeTagger object, and show the UI
if __name__ == '__main__':
    app = QApplication(sys.argv)

    tagger = createTimeTagger()

    # If you want to include this window within a bigger UI,
    # just copy these two lines within any of your handlers.
    window = CoincidenceExample(tagger)
    window.show()

    app.exec_()

    freeTimeTagger(tagger)



from __future__ import annotations

import json
import time
from datetime import datetime
from pathlib import Path

import numpy as np
import tifffile


def acquire_raw_flim(
    x_pixels,
    y_pixels,
    n_frames,
    pixel_dwell_ps,
    extra_left,
    extra_right,
    laser_frequency_hz,
    laser_ch,
    detector_ch,
    pixel_ch,
    laser_trigger_v,
    detector_trigger_v,
    pixel_trigger_v,
    laser_input_delay_ps=0,
    detector_input_delay_ps=0,
    pixel_input_delay_ps=0,
    laser_event_divider=1,
    use_conditional_filter=False,
    stream_buffer_size=4_000_000,
    poll_sleep_s=1e-4,
    final_pixel_margin_s=1e-3,
):
    from Swabian import TimeTagger

    if x_pixels <= 0 or y_pixels <= 0:
        raise ValueError("x_pixels and y_pixels must be positive.")
    if extra_left < 0 or extra_right < 0:
        raise ValueError("extra_left and extra_right must be non-negative.")
    if n_frames <= 0:
        raise ValueError("n_frames must be positive.")
    if pixel_dwell_ps <= 0:
        raise ValueError("pixel_dwell_ps must be positive.")
    if laser_frequency_hz <= 0:
        raise ValueError("laser_frequency_hz must be positive.")
    if not 1 <= int(laser_event_divider) <= 65535:
        raise ValueError("laser_event_divider must be in the range [1, 65535].")

    total_x_pixels = int(x_pixels) + int(extra_left) + int(extra_right)
    total_pixels = total_x_pixels * int(y_pixels)
    frame_duration_ps = int(total_pixels * int(pixel_dwell_ps))

    def new_pixel_lists():
        return [[] for _ in range(total_pixels)]

    def finalize_frame(pixel_lists):
        frame = np.empty((y_pixels, total_x_pixels), dtype=object)
        for flat_index, delays in enumerate(pixel_lists):
            y_index, x_index = divmod(flat_index, total_x_pixels)
            frame[y_index, x_index] = np.asarray(delays, dtype=np.int64)
        return frame[:, extra_left:extra_left + x_pixels]

    tagger = TimeTagger.createTimeTagger()
    stream = None
    frames = []
    stats = {
        "frames_started": 0,
        "laser_tags_seen": 0,
        "detector_tags_seen": 0,
        "pixel_tags_seen": 0,
        "detector_tags_assigned": 0,
        "detector_tags_without_laser": 0,
        "overflow_begin": 0,
        "overflow_end": 0,
        "missed_events": 0,
        "error_events": 0,
        "buffer_chunks_full": 0,
    }

    armed = False
    pixel_lists = None
    frame_start_ps = None
    frame_stop_ps = None
    last_laser_time_ps = None
    finalize_deadline = None

    try:
        tagger.setTriggerLevel(laser_ch, laser_trigger_v)
        tagger.setTriggerLevel(detector_ch, detector_trigger_v)
        tagger.setTriggerLevel(pixel_ch, pixel_trigger_v)

        adjusted_laser_delay_ps = int(laser_input_delay_ps)
        if use_conditional_filter:
            # This may distort the timing relationship between laser and detector
            # tags in the raw stream. Disable if the hardware stream looks wrong.
            tagger.setConditionalFilter(trigger=[detector_ch], filtered=[laser_ch])
        if int(laser_event_divider) > 1:
            tagger.setEventDivider(laser_ch, int(laser_event_divider))

        if adjusted_laser_delay_ps:
            tagger.setInputDelay(laser_ch, adjusted_laser_delay_ps)
        if detector_input_delay_ps:
            tagger.setInputDelay(detector_ch, int(detector_input_delay_ps))
        if pixel_input_delay_ps:
            tagger.setInputDelay(pixel_ch, int(pixel_input_delay_ps))

        stream = TimeTagger.TimeTagStream(
            tagger=tagger,
            n_max_events=stream_buffer_size,
            channels=[laser_ch, detector_ch, pixel_ch],                    # pyright:ignore
        )
        stream.start()

        while len(frames) < n_frames:
            data = stream.getData()

            if data.size == stream_buffer_size:
                stats["buffer_chunks_full"] += 1

            if data.size > 0:
                event_types = data.getEventTypes()
                non_time_mask = event_types != 0
                if np.any(non_time_mask):
                    stats["overflow_begin"] += int(np.count_nonzero(event_types == 2))
                    stats["overflow_end"] += int(np.count_nonzero(event_types == 3))
                    stats["error_events"] += int(np.count_nonzero(event_types == 1))
                    missed_events = data.getMissedEvents()
                    stats["missed_events"] += int(np.sum(missed_events[event_types == 4]))

                valid_mask = event_types == 0
                if np.any(valid_mask):
                    timestamps = data.getTimestamps()[valid_mask].astype(np.int64, copy=False)
                    channels = data.getChannels()[valid_mask].astype(np.int64, copy=False)

                    if not armed:
                        pixel_positions = np.flatnonzero(channels == pixel_ch)
                        if pixel_positions.size:
                            arm_position = int(pixel_positions[0])
                            pre_arm_lasers = timestamps[:arm_position][channels[:arm_position] == laser_ch]
                            if pre_arm_lasers.size:
                                last_laser_time_ps = int(pre_arm_lasers[-1])

                            timestamps = timestamps[arm_position:]
                            channels = channels[arm_position:]
                            armed = True
                            pixel_lists = new_pixel_lists()
                            frame_start_ps = int(timestamps[0])
                            frame_stop_ps = frame_start_ps + frame_duration_ps
                            finalize_deadline = time.monotonic() + (
                                frame_duration_ps / 1e12
                            ) + final_pixel_margin_s
                            stats["frames_started"] += 1

                    if armed and timestamps.size:
                        laser_ts = timestamps[channels == laser_ch]
                        detector_ts = timestamps[channels == detector_ch]
                        pixel_ts_all = timestamps[channels == pixel_ch]

                        stats["laser_tags_seen"] += int(laser_ts.size)
                        stats["detector_tags_seen"] += int(detector_ts.size)
                        stats["pixel_tags_seen"] += int(pixel_ts_all.size)

                        if frame_stop_ps is not None:
                            laser_ts = laser_ts[laser_ts < frame_stop_ps]
                            detector_ts = detector_ts[detector_ts < frame_stop_ps]

                        if detector_ts.size:
                            if laser_ts.size:
                                if last_laser_time_ps is None:
                                    laser_reference = laser_ts
                                else:
                                    laser_reference = np.empty(laser_ts.size + 1, dtype=np.int64)
                                    laser_reference[0] = last_laser_time_ps
                                    laser_reference[1:] = laser_ts

                                laser_indices = np.searchsorted(laser_reference, detector_ts, side="right") - 1
                                valid_detector_mask = laser_indices >= 0
                                valid_detector_ts = detector_ts[valid_detector_mask]
                                detector_delays_ps = (
                                    valid_detector_ts - laser_reference[laser_indices[valid_detector_mask]]
                                )
                            elif last_laser_time_ps is not None:
                                valid_detector_ts = detector_ts
                                detector_delays_ps = detector_ts - last_laser_time_ps
                                valid_detector_mask = np.ones(detector_ts.shape, dtype=bool)
                            else:
                                valid_detector_ts = detector_ts[:0]
                                detector_delays_ps = detector_ts[:0]
                                valid_detector_mask = np.zeros(detector_ts.shape, dtype=bool)

                            stats["detector_tags_without_laser"] += int(
                                detector_ts.size - np.count_nonzero(valid_detector_mask)
                            )

                            if valid_detector_ts.size:
                                detector_pixel_indices = (
                                    (valid_detector_ts - frame_start_ps) // int(pixel_dwell_ps)
                                ).astype(np.int64, copy=False)
                                in_frame_mask = (
                                    (detector_pixel_indices >= 0)
                                    & (detector_pixel_indices < total_pixels)
                                )

                                if np.any(in_frame_mask):
                                    assigned_pixels = detector_pixel_indices[in_frame_mask]
                                    assigned_delays = detector_delays_ps[in_frame_mask]
                                    order = np.argsort(assigned_pixels, kind="stable")
                                    assigned_pixels = assigned_pixels[order]
                                    assigned_delays = assigned_delays[order]
                                    split_points = np.flatnonzero(np.diff(assigned_pixels)) + 1

                                    for pixel_index_group, delay_group in zip(
                                        np.split(assigned_pixels, split_points),
                                        np.split(assigned_delays, split_points),
                                    ):
                                        pixel_lists[int(pixel_index_group[0])].extend(
                                            delay_group.tolist()
                                        )

                                    stats["detector_tags_assigned"] += int(assigned_delays.size)

                        if laser_ts.size:
                            last_laser_time_ps = int(laser_ts[-1])

                        crossed_frame_end = bool(np.any(timestamps >= frame_stop_ps))
                        if crossed_frame_end or time.monotonic() >= finalize_deadline:
                            frames.append(finalize_frame(pixel_lists))
                            armed = False
                            pixel_lists = None
                            frame_start_ps = None
                            frame_stop_ps = None
                            last_laser_time_ps = None
                            finalize_deadline = None

                            if len(frames) >= n_frames:
                                break

            if armed and finalize_deadline is not None and time.monotonic() >= finalize_deadline:
                frames.append(finalize_frame(pixel_lists))
                armed = False
                pixel_lists = None
                frame_start_ps = None
                frame_stop_ps = None
                last_laser_time_ps = None
                finalize_deadline = None

                if len(frames) >= n_frames:
                    break

            if data.size == 0:
                time.sleep(poll_sleep_s)

        return frames, stats

    finally:
        if stream is not None:
            try:
                stream.stop()
            except Exception:
                pass
        TimeTagger.freeTimeTagger(tagger)


def save_raw_flim_data(frames, output_dir, run_name, acquisition_parameters):
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    stem = output_dir / f"{timestamp}_{run_name}"
    frame_stack = np.asarray(frames, dtype=object)
    laser_period_ps = int(round(1e12 / acquisition_parameters["laser_frequency_hz"]))

    intensity = np.zeros((len(frame_stack), frame_stack.shape[1], frame_stack.shape[2]), dtype=np.uint32)
    mean_delay_ps = np.full((len(frame_stack), frame_stack.shape[1], frame_stack.shape[2]), np.nan, dtype=np.float32)

    for frame_index, frame in enumerate(frame_stack):
        for y_index in range(frame.shape[0]):
            for x_index in range(frame.shape[1]):
                delays = frame[y_index, x_index]
                intensity[frame_index, y_index, x_index] = len(delays)
                if len(delays):
                    mean_delay_ps[frame_index, y_index, x_index] = float(
                        np.mean(np.mod(delays, laser_period_ps))
                    )

    np.savez_compressed(
        stem.with_name(f"{stem.name}_raw.npz"),
        frames=frame_stack,
        acquisition_parameters=np.asarray(acquisition_parameters, dtype=object),
    )
    tifffile.imwrite(stem.with_name(f"{stem.name}_intensity.tiff"), intensity)
    tifffile.imwrite(stem.with_name(f"{stem.name}_mean_delay_ps.tiff"), mean_delay_ps)
    stem.with_name(f"{stem.name}_metadata.json").write_text(
        json.dumps(
            {
                "timestamp": datetime.now().isoformat(),
                "run_name": run_name,
                "acquisition_parameters": acquisition_parameters,
            },
            indent=2,
        ),
        encoding="utf-8",
    )

    return stem


if __name__ == "__main__":
    run_name = "flim_test_001"
    output_dir = Path("data") / "swabian_raw_flim"

    x_pixels = 512
    y_pixels = 512
    extra_left = 30
    extra_right = 10
    n_frames = 1
    pixel_dwell_ps = int(10e6)
    laser_frequency_hz = 80e6

    laser_ch = 1
    detector_ch = 2
    pixel_ch = 3

    laser_trigger_v = 0.05
    detector_trigger_v = 0.2
    pixel_trigger_v = 0.2

    laser_input_delay_ps = 0
    detector_input_delay_ps = 0
    pixel_input_delay_ps = 0
    laser_event_divider = 100
    use_conditional_filter = False

    acquisition_parameters = {
        "x_pixels": x_pixels,
        "y_pixels": y_pixels,
        "extra_left": extra_left,
        "extra_right": extra_right,
        "n_frames": n_frames,
        "pixel_dwell_ps": pixel_dwell_ps,
        "laser_frequency_hz": laser_frequency_hz,
        "laser_ch": laser_ch,
        "detector_ch": detector_ch,
        "pixel_ch": pixel_ch,
        "laser_trigger_v": laser_trigger_v,
        "detector_trigger_v": detector_trigger_v,
        "pixel_trigger_v": pixel_trigger_v,
        "laser_input_delay_ps": laser_input_delay_ps,
        "detector_input_delay_ps": detector_input_delay_ps,
        "pixel_input_delay_ps": pixel_input_delay_ps,
        "laser_event_divider": laser_event_divider,
        "use_conditional_filter": use_conditional_filter,
    }

    print("Starting raw FLIM acquisition.")
    print("Run this script first, then start the DAQ-side scan.")

    start_time = time.perf_counter()
    frames, stats = acquire_raw_flim(
        x_pixels=x_pixels,
        y_pixels=y_pixels,
        n_frames=n_frames,
        pixel_dwell_ps=pixel_dwell_ps,
        extra_left=extra_left,
        extra_right=extra_right,
        laser_frequency_hz=laser_frequency_hz,
        laser_ch=laser_ch,
        detector_ch=detector_ch,
        pixel_ch=pixel_ch,
        laser_trigger_v=laser_trigger_v,
        detector_trigger_v=detector_trigger_v,
        pixel_trigger_v=pixel_trigger_v,
        laser_input_delay_ps=laser_input_delay_ps,
        detector_input_delay_ps=detector_input_delay_ps,
        pixel_input_delay_ps=pixel_input_delay_ps,
        laser_event_divider=laser_event_divider,
        use_conditional_filter=use_conditional_filter,
    )
    elapsed_s = time.perf_counter() - start_time

    stem = save_raw_flim_data(
        frames=frames,
        output_dir=output_dir,
        run_name=run_name,
        acquisition_parameters=acquisition_parameters,
    )

    print(f"Finished in {elapsed_s:.3f} s")
    print(json.dumps(stats, indent=2))
    print(f"Saved raw data to {stem.with_name(f'{stem.name}_raw.npz')}")



"""In this example, we learn how to remove events on the device containing
no information to save USB bandwidth."""

from matplotlib import pyplot as plt
import sys

from Swabian import TimeTagger


def find_random_signal(tt, threshold_rate):
    # This function scans the trigger level in the same way as
    # it is shown in example 2-B.
    channels = tt.getChannelList(TimeTagger.ChannelEdge.Rising)
    level = max(-0.01, tt.getTriggerLevelRange(channels[0])[0])
    # If using the Time Tagger X, the input hysteresis value should be set to the lowest possible value for all input channels.
    if tagger.getModel() == 'Time Tagger X':
        [tt.setInputHysteresis(channel,  1) for channel in channels]
    count = TimeTagger.Countrate(tagger=tt, channels=channels)
    print("")
    while level < 0.01:
        sys.stdout.write("Searching level: {:5.1f} mV\r".format(level * 1000))
        for channel in channels:
            tt.setTriggerLevel(channel, level)
        count.startFor(int(1E8))
        count.waitUntilFinished()
        data = count.getData()
        if max(data) > threshold_rate:
            return channels[data.argmax()], level, True
        level += 0.0001
    return 0, .0, False


# Create a TimeTagger instance to control your hardware
tagger = TimeTagger.createTimeTagger()

# Set a threshold for the random counts according to the Time Tagger model
min_random_rate = 3E7 if tagger.getModel() == "Time Tagger 20" else 1E8

# Try to find a channel that exceeds the count rate threshold and set the
# trigger level accordingly
random_channel, random_level, random_found = find_random_signal(
    tagger, min_random_rate)
if random_found:
    print("\nFound a random signal on channel {} at {:2.3} mV.".format(
        random_channel, 1000 * random_level))
    tagger.setTriggerLevel(channel=random_channel, voltage=random_level)

    # The default dead-time is defined by the internal clock of the Time Tagger
    # TT X: 750 MHz, TT Ultra: 500 MHz, TT 20: 166.6 MHz
    default_deadtime = tagger.getDeadtime(random_channel)

    # To measure the count rate that exceeds the USB transfer limit without overflows,
    # we use the EventDivider. With divider=100, we discard 99 out of 100 events and
    # let only one pass every 100th event, which is within the USB bandwidth
    tagger.setEventDivider(channel=random_channel, divider=100)
    count = TimeTagger.Countrate(tagger, [random_channel])
    count.startFor(int(1E12))
    count.waitUntilFinished()

    # To obtain the real count rate, we have to multiply by 100
    countrate = count.getData()[0] * 100
    specified_usb_rate = 7.5E6 if tagger.getModel() == "Time Tagger 20" else 82E6
    print(
        "\nThe countrate of the random channel is {:.1f} Mtags/s".format(countrate / 1E6))
    print("This is {:.1f} % of the specified USB data rate and {:.1f} % of the maximal time-to-digital conversion rate.".format(
        100*countrate/specified_usb_rate, 100*countrate/(1E12/default_deadtime)
    ))

    # Switch off the EventDivider = set it to 1
    tagger.setEventDivider(channel=random_channel, divider=1)

    # Take a second input channel which is NOT the random channel and let it
    # measure the built-in test signal, i.e. a periodic signal
    periodic_channel = 2 if random_channel == 1 else 1
    tagger.setTestSignal(periodic_channel, True)

    # We want to measure only the very first random click after a periodic click.
    # This is achieved by the ConditionalFilter. It opens a gate for any event
    # on the 'trigger' channels and lets only one event of each 'filtered' channel
    # pass. After that, the gate is closed until the next trigger event.
    tagger.setConditionalFilter(
        trigger=[periodic_channel], filtered=[random_channel])

    # We measure the correlation between the periodic and the FILTERED random channels
    corr = TimeTagger.Correlation(tagger=tagger,
                                  channel_1=random_channel,
                                  channel_2=periodic_channel,
                                  binwidth=default_deadtime//100,
                                  n_bins=3000)
    plt.figure(figsize=(9, 6))
    print("\nMeasuring with different dead-times:")
    for i in range(1, 6):
        # We repeat this measurement for five different dead-times
        tagger.setDeadtime(random_channel, i * default_deadtime)
        deadtime = tagger.getDeadtime(random_channel)
        print("Default dead-time x {} = {}".format(i, deadtime))
        corr.startFor(int(1E12))
        corr.waitUntilFinished()
        plt.plot(corr.getIndex()/default_deadtime,
                 corr.getData(), label="{} ps".format(deadtime))
    plt.title("Correlation for different dead-times")
    plt.xlabel("Delay/(default_deadtime = {} ps)".format(default_deadtime))
    plt.ylabel("Counts/bin")
    plt.annotate("""Within the dead-time,
the Correlation is more
or less constant, i.e
randomly distributed.
For larger delays, the
Correlation drops with
the probability of finding
no preceding time-tag.""",
                 (2, max(corr.getData())),
                 xytext=(-150, 0),
                 textcoords='offset pixels',
                 ha='right',
                 arrowprops={'arrowstyle': '->'})
    plt.legend()
    plt.show()

TimeTagger.freeTimeTagger(tagger)



"""
Use this script to measure calibration time tags.
You will need to adjust the settings in pnr_constants.py file to your specific
configuration.
"""

from pnr_constants import FRONT_EDGE, REAR_EDGE, LASER_CHANNEL, FOLDER, FILENAME

from Swabian import TimeTagger


DURATION = 5E12

# Adjust these settings according to your setup
DETECTOR_TRIGGER_LEVEL = 0.5
LASER_TRIGGER_LEVEL = 0.5

# Setup the tagger, the ConditionalFilter ensures the sequence falling edge -> laser
tagger = TimeTagger.createTimeTagger()
tagger.setTriggerLevel(FRONT_EDGE, DETECTOR_TRIGGER_LEVEL)
tagger.setTriggerLevel(LASER_CHANNEL, LASER_TRIGGER_LEVEL)

# Run file dump for defined duration
sync = TimeTagger.SynchronizedMeasurements(tagger)
fw = TimeTagger.FileWriter(tagger=sync.getTagger(),
                           filename=FOLDER + FILENAME,
                           channels=[FRONT_EDGE, REAR_EDGE, LASER_CHANNEL])
sync.startFor(DURATION)
sync.waitUntilFinished()



from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np


@dataclass
class RPOCImageInput:
    data: np.ndarray
    channel_labels: list[str] = field(default_factory=list)
    source_id: str = ""

    def __post_init__(self) -> None:
        arr = np.asarray(self.data, dtype=np.float32)
        if arr.ndim != 3:
            raise ValueError("RPOCImageInput.data must be channels-first [C, H, W]")
        if arr.shape[0] <= 0:
            raise ValueError("RPOCImageInput.data must contain at least one channel")
        if arr.shape[1] <= 0 or arr.shape[2] <= 0:
            raise ValueError("RPOCImageInput.data has invalid spatial dimensions")
        self.data = arr
        if not self.channel_labels:
            self.channel_labels = [f"Channel {i + 1}" for i in range(arr.shape[0])]
        if len(self.channel_labels) != arr.shape[0]:
            raise ValueError("channel_labels length must match channel dimension")

    @classmethod
    def from_array(
        cls,
        image_data: np.ndarray,
        channel_labels: list[str] | None = None,
        source_id: str = "",
    ) -> "RPOCImageInput":
        arr = np.asarray(image_data, dtype=np.float32)
        if arr.ndim == 2:
            arr = arr[None, ...]
        elif arr.ndim == 3:
            if arr.shape[0] <= 8:
                pass
            elif arr.shape[-1] <= 8:
                arr = np.moveaxis(arr, -1, 0)
            else:
                raise ValueError(
                    "3D input shape is ambiguous; expected channels-first [C,H,W] "
                    "or channels-last [H,W,C] with C <= 8"
                )
        else:
            raise ValueError("RPOC input must be 2D or 3D data")
        labels = channel_labels or [f"Channel {i + 1}" for i in range(arr.shape[0])]
        return cls(data=arr, channel_labels=labels, source_id=source_id)


@dataclass
class RPOCRoi:
    roi_id: int
    points: list[tuple[float, float]]
    threshold_low: float
    threshold_high: float
    modulation_level: float = 0.5
    active_channels: list[bool] = field(default_factory=list)


@dataclass
class RPOCEditorState:
    image_input: RPOCImageInput | None = None
    channel_visibility: list[bool] = field(default_factory=list)
    rois: list[RPOCRoi] = field(default_factory=list)
    show_rois: bool = True
    show_labels: bool = True
    next_roi_id: int = 1



"""If errors occur during the operation of the Time Tagger, for example, due to the loss of the USB connection,
the Time Tagger notifies you. You can register a callback function to handle errors in the way
you like."""

from time import sleep

from Swabian import TimeTagger

# In general, we can register any callable object as logger callback, so a simple function like
#
# def log_handler(level, msg):
#     print(msg)
#
# would suffice. Here, however, we define a callable class instance to keep track of the number
# of messages that have been sent.


class LogHandler:
    """A callable class that also keeps a state."""

    def __init__(self):
        # We want to count the number of messages to stop the program after two messages
        self.number_of_messages = 0

    def __call__(self, level, msg):
        # The __call__ method makes our class instance callable

        print("")
        # level represents different types of messages
        if level == TimeTagger.LOGGER_ERROR:
            print("An error occurred:")
        elif level == TimeTagger.LOGGER_WARNING:
            print("This is a Time Tagger warning:")
        elif level == TimeTagger.LOGGER_INFO:
            print("Just for your information:")
        else:
            print("This should not have happened...")
        print(msg)
        self.number_of_messages += 1


# Now we create an instance of the LogHandler class and register it as the logger callback
log_handler = LogHandler()
TimeTagger.setLogger(callback=log_handler)

# Create a TimeTagger instance to control your hardware
tagger = TimeTagger.createTimeTagger()

# To test our logger, we create an error by disconnecting the Time Tagger from the PC.
# If you re-plug it, the Time Tagger sends an info notification.
print("\n-> Please unplug the USB of your Time Tagger and reconnect it afterward.")
while log_handler.number_of_messages < 2:
    sleep(.1)
print("\nMission accomplished!")

TimeTagger.freeTimeTagger(tagger)



from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class Action:
    label: str
    method_name: str
    parameters: list["BaseParameter"] = field(default_factory=list)
    tooltip: str = ""
    dangerous: bool = False
    confirm_text: str | None = None


ParameterGroups = dict[str, list["BaseParameter"]]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QListWidgetItem, QMessageBox

from pyrpoc.displays.display_registry import display_registry
from pyrpoc.gui.main_widgets.display_mgr.forms import prompt_display_parameters

if TYPE_CHECKING:
    from pyrpoc.gui.main_widgets.display_mgr.widget import DisplayManagerWidget


def refresh_available(widget: DisplayManagerWidget) -> None:
    """
    Build display dropdown constrained by selected modality contract.

    Route:
    - modality selected signal
    - -> this handler
    - -> display list filtered by output type and allowed display keys.
    """
    contract = widget.modality_service.get_selected_contract()
    output_contract = contract.get("output_data_contract")
    allowed_displays = set(contract.get("allowed_displays", []))
    current_key = widget._selected_display_key()
    available_rows = {row["key"]: row for row in widget.display_service.list_available()}

    widget.display_combo.clear()
    if isinstance(output_contract, str) and output_contract.strip():
        keys = widget.display_service.list_compatible_with(output_contract.strip())
    else:
        keys = list(available_rows.keys())

    if allowed_displays:
        keys = [key for key in keys if key in allowed_displays]

    for key in keys:
        row = available_rows.get(key, {"display_name": key})
        widget.display_combo.addItem(row.get("display_name", key), key)

    if current_key:
        idx = widget.display_combo.findData(current_key)
        if idx >= 0:
            widget.display_combo.setCurrentIndex(idx)


def refresh_instances(widget: DisplayManagerWidget) -> None:
    """
    Rebuild visible instance list from service inventory rows.

    Route:
    - display add/remove/attach/detach signal
    - -> this handler
    - -> list widget rows with stable object identity in UserRole.
    """
    current = widget._selected_display()
    current_id = id(current) if current is not None else None
    widget.instances_list.blockSignals(True)
    try:
        widget.instances_list.clear()
        for idx, row in enumerate(widget.display_service.list_instances(), start=1):
            raw_display_id = row.get("display_id")
            display_id: int | None = None
            if isinstance(raw_display_id, int):
                display_id = raw_display_id
            elif isinstance(raw_display_id, bool):
                display_id = int(raw_display_id)
            else:
                state_value = row.get("state")
                if isinstance(state_value, int):
                    display_id = state_value
                elif state_value is not None:
                    try:
                        display_id = id(state_value)
                    except Exception:
                        display_id = None

            if display_id is None:
                continue
            marker = "attached" if row["attached"] else "detached"
            item = QListWidgetItem(f"{row['name']} [{idx}] ({marker})")
            item.setFlags(item.flags() | Qt.ItemFlag.ItemIsEditable)
            item.setData(Qt.ItemDataRole.UserRole, display_id)
            widget.instances_list.addItem(item)

        if current_id is not None:
            for idx in range(widget.instances_list.count()):
                item = widget.instances_list.item(idx)
                if item.data(Qt.ItemDataRole.UserRole) == current_id:
                    widget.instances_list.setCurrentRow(idx)
                    break
    finally:
        widget.instances_list.blockSignals(False)


def on_display_name_edited(widget: DisplayManagerWidget, item: Any) -> None:
    if not isinstance(item, QListWidgetItem):
        return
    raw = item.text().strip()
    if not raw:
        return
    display_id = item.data(Qt.ItemDataRole.UserRole)
    if not isinstance(display_id, int):
        return
    display = widget.display_service.get_display_by_id(display_id)
    if display is None:
        return

    name = raw.split(" [", 1)[0].strip()
    if not name:
        return
    widget.display_service.set_display_name(display, name)


def on_add_clicked(widget: DisplayManagerWidget) -> None:
    """
    Add one display using the selected registry key and optional parameter dialog.

    Route:
    - Add button click
    - -> this handler
    - -> `DisplayService.create_display`
    - -> `display_added` tab/list refresh.
    """
    key = widget._selected_display_key()
    if not key:
        show_error(widget, "No compatible display available for selected modality")
        return

    raw_settings: dict[str, Any] = {}
    display_cls = display_registry.get_class(key)
    display_name = widget.name_input.text().strip()
    if not display_name:
        row = widget.display_service.list_available()
        fallback = next((row_entry for row_entry in row if row_entry["key"] == key), None)
        display_name = str(fallback.get("display_name", key)) if isinstance(fallback, dict) else key

    if any(display_cls.DISPLAY_PARAMETERS.values()):
        settings = prompt_display_parameters(widget, display_cls.DISPLAY_PARAMETERS)
        if settings is None:
            return
        raw_settings = settings

    try:
        widget.display_service.create_display(key, raw_settings, user_label=display_name)
        widget.status_label.setText("Status: added display")
        widget.name_input.clear()
    except Exception as exc:
        show_error(widget, str(exc))


def on_attach_clicked(widget: DisplayManagerWidget) -> None:
    state = widget._selected_display()
    if state is None:
        return
    widget.display_service.attach(state)
    widget.status_label.setText("Status: attached display")
    refresh_instances(widget)


def on_detach_clicked(widget: DisplayManagerWidget) -> None:
    state = widget._selected_display()
    if state is None:
        return
    widget.display_service.detach(state)
    widget.status_label.setText("Status: detached display")
    refresh_instances(widget)


def on_remove_clicked(widget: DisplayManagerWidget) -> None:
    state = widget._selected_display()
    if state is None:
        return
    widget.display_service.remove_display(state)
    widget.status_label.setText("Status: removed display")


def on_display_changed(widget: DisplayManagerWidget, state: object) -> None:
    del state
    widget.status_label.setText("Status: display updated")
    refresh_instances(widget)


def on_display_removed(widget: DisplayManagerWidget, state: object) -> None:
    del state
    refresh_instances(widget)


def on_display_error(widget: DisplayManagerWidget, _state: object, message: str) -> None:
    widget.status_label.setText(f"Status: display error - {message}")


def on_modality_selected(widget: DisplayManagerWidget, key: str) -> None:
    del key
    refresh_available(widget)


def show_error(widget: DisplayManagerWidget, message: str) -> None:
    widget.status_label.setText(f"Status: error - {message}")
    QMessageBox.critical(widget, "Display Error", message)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from PyQt6.QtWidgets import QWidget

from pyrpoc.backend_utils.array_contracts import CONTRACT_CHW_FLOAT32
from pyrpoc.backend_utils.contracts import ParameterGroups
from pyrpoc.backend_utils.parameter_utils import validate_parameter_groups
from pyrpoc.backend_utils.state_helpers import export_object_state, import_object_state, make_instance_id
from pyrpoc.rpoc.types import RPOCImageInput

if TYPE_CHECKING:
    from pyrpoc.domain.app_state import ParameterValue
    from collections.abc import Callable


class BaseDisplay(QWidget):
    DISPLAY_KEY: str = "base_display"
    DISPLAY_NAME: str = "Base Display"
    ACCEPTED_DATA_CONTRACTS: list[str] = [CONTRACT_CHW_FLOAT32]
    DISPLAY_PARAMETERS: ParameterGroups = {}
    PERSISTENCE_FIELDS: tuple[str, ...] | None = None
    PERSISTENCE_EXCLUDE_FIELDS: tuple[str, ...] = (
        "attached",
        "config_values",
        "docked_visible",
        "instance_id",
        "last_error",
        "user_label",
    )

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        accepted = getattr(cls, "ACCEPTED_DATA_CONTRACTS", [])
        if not isinstance(accepted, list):
            raise TypeError("ACCEPTED_DATA_CONTRACTS must be a list")
        for contract in accepted:
            if not isinstance(contract, str) or not contract.strip():
                raise TypeError("ACCEPTED_DATA_CONTRACTS must contain non-empty strings")

        validate_parameter_groups(getattr(cls, "DISPLAY_PARAMETERS", {}))

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        # Runtime metadata used by DisplayService/session persistence.
        self.attached: bool = True
        self.docked_visible: bool = True
        self.config_values: list[ParameterValue] = []
        self.user_label: str | None = None
        self.instance_id: str = make_instance_id(self.DISPLAY_KEY)
        self.last_error: str | None = None
        self._persist_callback: Callable[[], None] | None = None

    @property
    def type_key(self) -> str:
        """Registry key used by session restore and inventory rows."""
        return self.DISPLAY_KEY

    @classmethod
    def get_contract(cls) -> dict[str, Any]:
        return {
            "display_key": cls.DISPLAY_KEY,
            "display_name": cls.DISPLAY_NAME,
            "accepted_data_contracts": cls.ACCEPTED_DATA_CONTRACTS,
            "display_parameters": cls.DISPLAY_PARAMETERS,
        }

    def configure(self, params: dict[str, Any]) -> None:
        raise NotImplementedError

    def render(self, data: np.ndarray) -> None:
        raise NotImplementedError

    def clear(self) -> None:
        raise NotImplementedError

    def export_rpoc_input(self) -> RPOCImageInput | None:
        return None

    def get_normalized_data_3d(self) -> np.ndarray | None:
        """Return current display data as float32 [C,H,W] in [0,1], or None if unavailable."""
        return None

    def export_persistence_state(self) -> dict[str, Any]:
        return export_object_state(
            self,
            include_fields=self.PERSISTENCE_FIELDS,
            exclude_fields=self.PERSISTENCE_EXCLUDE_FIELDS,
        )

    def import_persistence_state(self, state: dict[str, Any]) -> None:
        import_object_state(
            self,
            state,
            include_fields=self.PERSISTENCE_FIELDS,
            exclude_fields=self.PERSISTENCE_EXCLUDE_FIELDS,
        )

    def set_persist_callback(self, callback: "Callable[[], None] | None") -> None:
        self._persist_callback = callback

    def request_persist(self) -> None:
        if self._persist_callback is not None:
            self._persist_callback()



from custom_pnr_constants import FRONT_EDGE, REAR_EDGE, LASER_CHANNEL, FOLDER, FILENAME, CALIBRATION_FILES
from json import load
import numba
import numpy as np
import os

from Swabian import TimeTagger


class CustomPNRMultichannel(TimeTagger.CustomMeasurement):
    """
    Custom measurement to perform photon number resolution on multiple detectors simultaneously using one common trigger signal.
    Input parameters, i.e., trigger and detector channels, are imported from an external file named custom_pnr_constants.py. 
    The data can be accessed using the getData() module. The output is a numpy array with one row for each time tag and a column for each detector, storing the number of detected photons.
    """

    def __init__(self, tagger, trigger_ch, pnr_channels, n_detectors, n_channels, buffer_size):
        TimeTagger.CustomMeasurement.__init__(self, tagger)
        self.trigger_ch = trigger_ch
        self.pnr_channels = pnr_channels
        self.n_detectors = n_detectors
        self.n_channels = n_channels
        self.buffer_size = int(buffer_size)

        # Channels that will be transferred from the TT to the PC
        self.register_channel(channel=trigger_ch)
        for pnr_channel in np.nditer(pnr_channels):
            self.register_channel(channel=pnr_channel)

        self.clear_impl()

        self.finalize_init()

    def __del__(self):
        # The measurement must be stopped before deconstruction to avoid
        # concurrent process() calls.
        self.stop()

    def getData(self):
        # Acquire a lock this instance to guarantee that process() is not running in parallel
        # This ensures to return a consistent data.
        with self.mutex:
            # Check if there are new data. if not return an array of -1
            if self.index == self.read_index:
                return np.zeros((0, self.n_detectors), dtype=np.int8)
            else:
                if not self.state['is_ring_full']:
                    data = self.data[self.read_index:self.index].copy()
                    self.read_index = self.index
                else:
                    data = np.concatenate(
                        (self.data[self.index:], self.data[self.read_index:self.index]))
                    self.read_index = self.index
            return data

    def clear_impl(self):
        # The lock is already acquired within the backend.
        self.data = np.full(
            (self.buffer_size, self.n_detectors), fill_value=-1, dtype=np.int8)
        self.index = 0
        self.read_index = 0
        self.timestamp = 0
        self.state = np.array([(False, False)], dtype=[
                              ('has_laser', bool), ('is_ring_full', bool)])

    def on_start(self):
        # The lock is already acquired within the backend.
        pass

    def on_stop(self):
        # The lock is already acquired within the backend.
        pass

    @staticmethod
    @numba.jit(nopython=True, nogil=True)
    def fast_process(
            tags,
            timestamp,
            data,
            trigger_ch,
            pnr_channels,
            n_detectors,
            n_channels,
            index,
            read_index,
            buffer,
            state):
        data[0][:] = -1
        for tag in tags:
            # print(tag['type'], tag['time'], tag['channel'])
            # tag.type can be: 0 - TimeTag, 1- Error, 2 - OverflowBegin, 3 -
            # OverflowEnd, 4 - MissedEvents (you can use the TimeTagger.TagType IntEnum)
            if tag['type'] != TimeTagger.TagType.TimeTag:
                # tag is not a TimeTag, so we are in an error state, e.g. overflow
                continue
            elif tag['channel'] == trigger_ch:
                if timestamp == 0:
                    # This is a new timestamp. It starts a new event
                    timestamp = tag['time']
                    state[0]['has_laser'] = True
                elif timestamp == tag['time']:
                    # It is still the same event
                    if state[0]['has_laser']:
                        # This is a sanity check. There shouldn't be two laser events with the same timestamp
                        raise RuntimeError(
                            'Two Laser events with the same timestamp')
                    else:
                        state[0]['has_laser'] = True
                else:
                    # This is a new timestamp. Check if the laser was missing and in case keep the same index
                    if not state[0]['has_laser']:
                        pass
                    # Otherwise increase the index by one
                    else:
                        # First check if the buffer will be exceeded by doing it and data will be lost
                        if (index + 1 == read_index) or (index + 1 == buffer and read_index == 0):
                            raise RuntimeError(
                                'Buffer exceeded! Please increase the buffer, or call getData() more ofter')
                        # If there is still space, increase the index
                        index += 1
                        # Check if ring is full, and in case start from the beginning. Data in the first rows were saved already
                        if index == buffer:
                            index = 0
                            state[0]['is_ring_full'] = True
                    # assign the new time stamp
                    timestamp = tag['time']
                    state[0]['has_laser'] = True
            else:
                idx = 0
                for detector_index in range(n_detectors):
                    for channel_index in range(n_channels[detector_index]):
                        if tag['channel'] == pnr_channels[idx]:
                            if timestamp == 0:
                                # This is a new timestamp. It starts a new event
                                timestamp = tag['time']
                                #  And write the number of photons in the right column
                                data[index][detector_index] = channel_index
                            elif timestamp == tag['time']:
                                # It is still the same event. Write the number of photons in the right column
                                data[index][detector_index] = channel_index
                            else:
                                # This is a new event. Check if the laser was missing and in case keep the same index
                                if not state[0]['has_laser']:
                                    pass
                                # Otherwise increase the index by one checking if it fills the buffer
                                else:
                                    # First check if the buffer will be exceeded by doing it and data will be lost
                                    if (index + 1 == read_index) or (index + 1 == buffer and read_index == 0):
                                        raise RuntimeError(
                                            'Buffer exceeded! Please increase the buffer, or call getData() more ofter')
                                    # If there is still space, increase the index
                                    index += 1
                                    # Check if ring is full, and in case start from the beginning. Data in the first rows were saved already
                                    if index == buffer:
                                        index = 0
                                        state[0]['is_ring_full'] = True
                                # Assign the new time stamp
                                timestamp = tag['time']
                                # Set the laser state to false
                                state[0]['has_laser'] = False
                                # And write the number of photons in the right column
                                data[index][detector_index] = channel_index
                        idx += 1
        return index, read_index

    def process(self, incoming_tags, begin_time, end_time):
        self.index, self.read_index = CustomPNRMultichannel.fast_process(
            incoming_tags,
            self.timestamp,
            self.data,
            self.trigger_ch,
            self.pnr_channels,
            self.n_detectors,
            self.n_channels,
            self.index,
            self.read_index,
            self.buffer_size,
            self.state)


# Buffer size
BUFFER_SIZE = 1e6

if __name__ == '__main__':

    print("""Custom multichannel PNR measurement

""")

    if FRONT_EDGE and REAR_EDGE:
        if len(FRONT_EDGE) != len(REAR_EDGE):
            raise ValueError('The number of FRONT_EDGE and REAR_EDGE should be the same. There were given in input {} FRONT and {} REAR edges'.format(
                len(FRONT_EDGE), len(REAR_EDGE)))
        else:
            n_detectors = len(FRONT_EDGE)
    else:
        raise RuntimeError('FRONT and/or REAR edges not found')

    if len(CALIBRATION_FILES) != n_detectors:
        raise RuntimeError('{} calibration files given with {} detectors. Exactly one calibration file for each detector is needed.'.format(
            len(CALIBRATION_FILES), n_detectors))

    parameters: list[dict] = []
    n_channels = np.zeros((n_detectors), dtype=np.int8)
    # Looping over all calibration files
    for i, calib_file in enumerate(CALIBRATION_FILES):
        # if no absolute path is given, it assumed the calibration files to be in the same folder as the time tags data
        if os.path.basename(calib_file) == '':
            with open(FOLDER + calib_file, "rb") as param_file:
                parameters[i] = load(param_file)
        else:
            with open(FOLDER + calib_file, "rb") as param_file:
                parameters[i] = load(param_file)
        n_channels[i] = len(parameters[i]["x_intercepts"])

    ttv = TimeTagger.createTimeTaggerVirtual(FOLDER + FILENAME)

    pnr_measurements = []
    pnr_channels = np.empty(0, np.sum(n_channels))

    for i in range(n_detectors):

        print(i, parameters[i]["x_intercepts"])

        pnr_measurements.append(TimeTagger.Experimental.PhotonNumber(tagger=ttv,
                                                                     trigger_ch=LASER_CHANNEL,
                                                                     signal_start_ch=FRONT_EDGE[i],
                                                                     signal_stop_ch=REAR_EDGE[i],
                                                                     slope=parameters[i]["slope"],
                                                                     x_intercepts=parameters[i]["x_intercepts"],
                                                                     dead_time=100_000))
        pnr_channels = np.append(
            pnr_channels, pnr_measurements[i].getChannels())

    print(pnr_channels)

    multipnr = CustomPNRMultichannel(
        ttv, LASER_CHANNEL, pnr_channels, n_detectors, n_channels, BUFFER_SIZE)

    ttv.run()

    with open('pnr_counts.txt', 'w') as f:
        while not ttv.waitUntilFinished(timeout=1):
            np.savetxt(f, multipnr.getData(), fmt="%s")



"""
This is an example on how to connect and control the Time Tagger remotely using Pyro5.

See the tutorial article in the Time Tagger's Documentation:
https://www.swabianinstruments.com/static/documentation/TimeTagger/tutorials/TimeTaggerRPC.html

This example requires Pyro5 package that can be installed as:
    > pip install Pyro5

How to start this example:

    1. Start the server:
        > python server_example.py

    2. Run this example:
        > python client_example.py

"""
import base64
import io
import numpy as np
from matplotlib import pyplot as plt

try:
    import Pyro5.api
except ModuleNotFoundError:
    import sys
    print('Please install Pyro5 module. "python -m pip install Pyro5"')
    sys.exit()


def load_numpy_array(classname, data):
    assert classname == 'numpy.ndarray'
    buffer = io.BytesIO(base64.b64decode(data['data'].encode('ASCII')))
    return np.load(buffer, allow_pickle=False)


# Register numpy.ndarray deserializer with Pyro5
Pyro5.api.register_dict_to_class(classname='numpy.ndarray', converter=load_numpy_array)

TimeTagger = Pyro5.api.Proxy("PYRO:TimeTagger@localhost:23000")

# Create Time Tagger
tagger = TimeTagger.createTimeTagger()
tagger.setTestSignal(1, True)
tagger.setTestSignal(2, True)

print('Time Tagger serial:', tagger.getSerial())

hist = TimeTagger.Correlation(tagger, 1, 2, binwidth=2, n_bins=2000)
hist.startFor(int(10e12), clear=True)

fig, ax = plt.subplots()
# The time vector is fixed. No need to read it on every iteration.
x = hist.getIndex()
line, = ax.plot(x, x * 0)
ax.set_xlabel('Time (ps)')
ax.set_ylabel('Counts')
ax.set_title('Correlation histogram via Pyro-RPC')
while hist.isRunning():
    y = hist.getData()
    line.set_ydata(y)
    ax.set_ylim(np.min(y), np.max(y))
    plt.pause(0.1)

# Cleanup
TimeTagger.freeTimeTagger(tagger)
del hist
del tagger
del TimeTagger



from __future__ import annotations

from PyQt6.QtCore import pyqtSignal
from PyQt6.QtGui import QAction, QActionGroup
from PyQt6.QtWidgets import QMenu, QMenuBar
from PyQt6 import sip


class MainMenuBar(QMenuBar):
    style_selected = pyqtSignal(str)
    new_requested = pyqtSignal()
    open_requested = pyqtSignal()
    save_requested = pyqtSignal()
    save_as_requested = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.file_menu = QMenu("&File", self)
        self.new_action = QAction("New", self)
        self.new_action.triggered.connect(self.new_requested.emit)
        self.file_menu.addAction(self.new_action)

        self.open_action = QAction("Open...", self)
        self.open_action.triggered.connect(self.open_requested.emit)
        self.file_menu.addAction(self.open_action)

        self.save_action = QAction("Save", self)
        self.save_action.triggered.connect(self.save_requested.emit)
        self.file_menu.addAction(self.save_action)

        self.save_as_action = QAction("Save As...", self)
        self.save_as_action.triggered.connect(self.save_as_requested.emit)
        self.file_menu.addAction(self.save_as_action)

        self.file_menu.addSeparator()
        self.file_menu.addAction(QAction("Exit", self))
        self.addMenu(self.file_menu)

        self.view_menu = QMenu("View", self)
        self.addMenu(self.view_menu)

        self.prefs_menu = QMenu("&Preferences", self)
        self.prefs_menu.addAction(QAction("Application Settings", self))
        self.prefs_menu.addAction(QAction("Keyboard Shortcuts", self))
        self.addMenu(self.prefs_menu)

        self.style_menu = QMenu("&Style", self)
        self.addMenu(self.style_menu)
        self._style_actions: dict[str, QAction] = {}
        self._style_group = QActionGroup(self)
        self._style_group.setExclusive(True)

    def populate_view_menu(self, docks: list, display_actions: list[QAction] | None = None) -> None:
        self.view_menu.clear()
        for dock in docks:
            self.view_menu.addAction(dock.toggleViewAction())
        if display_actions:
            if docks:
                self.view_menu.addSeparator()
            for action in display_actions:
                if sip.isdeleted(action):
                    continue
                if not action.isCheckable():
                    action.setCheckable(True)
                self.view_menu.addAction(action)

    def populate_style_menu(self, selected_mode: str) -> None:
        self.style_menu.clear()
        self._style_actions.clear()
        mode_items = [
            ("system", "Follow System"),
            ("dark", "Dark"),
            ("light", "Light"),
        ]
        for mode, label in mode_items:
            action = QAction(label, self.style_menu)
            action.setCheckable(True)
            action.triggered.connect(lambda checked, m=mode: self.style_selected.emit(m))
            self._style_group.addAction(action)
            self.style_menu.addAction(action)
            self._style_actions[mode] = action
        self.set_active_style(selected_mode)

    def set_active_style(self, selected_mode: str) -> None:
        for mode, action in self._style_actions.items():
            action.setChecked(mode == selected_mode)
